/******************************** CPPFile *****************************

* FileName [AbsRefiner.cpp]

* PackageName [main]

* Synopsis [This file contains method definitions of the AbsRefiner
* class.]

* SeeAlso [AbsRefiner.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <map>
#include <vector>
#include <set>
#include <stack>
#include <typeinfo>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "Action.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "GlobalCEDag.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "CEDagVerifier.h"
#include "PredAbsRefiner.h"
#include "LtsAbsRefiner.h"
#include "AbsRefiner.h"

/*********************************************************************/
//define static members of AbsRefiner
/*********************************************************************/
const int AbsRefiner::CE_VALID = 11500;
const int AbsRefiner::CE_INVALID_REFINED = 11510;
const int AbsRefiner::CE_INVALID_NOT_REFINED = 11520;

/*********************************************************************/
//do it all. if the counter example is invalid add new predicates and
//return false. otherwise return true.
/*********************************************************************/
int AbsRefiner::Run()
{
  //compute the global CE dag
  GlobalCEDag::Initialize();
  Util::Message(2,"global counter example dag initialized ...\n");

  //flag to indicate if a valid CE was found
  bool validCE = false;
  //the index of the first component with an invalid dag
  int firstInvDag = -1;
  //the index of the first component with an invalid path
  int firstInvPath = -1;

  //clear all spurious concrete CEs
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    Database::procManagers[i].dagToLocs.clear();
    Database::procManagers[i].absLtsRefine.clear();
  }

  for(size_t j = 0;(!validCE) && (j < Database::MAX_GLOBAL_CE);j++) {
    //compute the global CE dag - continue only if this is a new dag
    if(!GlobalCEDag::Compute()) {
      assert(j > 0);
      break;
    }
    Util::Message(2,"global counter example dag computed ...\n");

    //do abstraction refinement on each component procedure
    for(size_t i = 0;i < Database::procManagers.size();i++) {
      int x = CEDagVerifier(Database::procManagers[i]).Run();
      if(x == CEDagVerifier::CE_VALID) {
	validCE = true;
	break;
      } else if(x == CEDagVerifier::CE_INVALID_DAG) {
	firstInvDag = (firstInvDag == -1) ? i : firstInvDag;
      } else if(x == CEDagVerifier::CE_INVALID_PATH) {
	firstInvPath = (firstInvPath == -1) ? i : firstInvPath;
      } else assert(false);
    }
    Util::Message(2,"CE dag projections analysed ...\n");
  }

  //if this counter example is valid
  if(validCE) return CE_VALID;
  //if this counter example is invalid because of lts abstraction
  else if(firstInvDag != -1) { 
    int res = LtsAbsRefiner(Database::procManagers[firstInvDag]).Run();
    assert(res == LtsAbsRefiner::LTS_ABS_REFINED);
    Database::lastRefType = Database::REFINE_LTS;
    Database::lastRefComp = firstInvDag;
    return CE_INVALID_REFINED;
  }
  //if this counter example is invalid because of predicate
  //abstraction
  else {
    assert(firstInvPath != -1);
    int res = PredAbsRefiner(Database::procManagers[firstInvPath]).Run();
    if(res == PredAbsRefiner::REFINED_NEW_PREDS) {
      Database::lastRefType = Database::REFINE_PRED;
      Database::lastRefComp = firstInvPath;
      return CE_INVALID_REFINED;
    } else {
      assert(res == PredAbsRefiner::REFINED_NO_NEW_PREDS);
      return CE_INVALID_NOT_REFINED;
    }
  }
}

/*********************************************************************/
//end of AbsRefiner.cpp
/*********************************************************************/
