/******************************** CPPFile *****************************

* FileName [ContLoc.cpp]

* PackageName [main]

* Synopsis [Method definitions of ContLoc class.]

* SeeAlso [ContLoc.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <set>
#include <map>
#include <vector>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "ProcAbs.h"
#include "Action.h"
#include "LtsTrans.h"
#include "FspInfo.h"
#include "ProcInfo.h"
#include "LtsInfo.h"
#include "LtlFormula.h"
#include "LtlManager.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ImplState.h"
#include "LocalAbsLts.h"
#include "ContLoc.h"
#ifndef MAGIC_LITE
#include "SAT.h"
#endif //MAGIC_LITE
#include "SatSimChecker.h"
#include "ConcCEDag.h"
#include "ProcManager.h"

/*********************************************************************/
//methods for ContLocBackup class
/*********************************************************************/
ContLocBackup::ContLocBackup() { tSucc = eSucc = callSucc = NULL; }

ContLocBackup::~ContLocBackup() {}

ContLocBackup::ContLocBackup(const ContLocBackup &rhs) { *this = rhs; }

const ContLocBackup &ContLocBackup::operator = (const ContLocBackup &rhs)
{
  tSucc = rhs.tSucc;
  eSucc = rhs.eSucc;
  callSucc = rhs.callSucc;
  preds = rhs.preds;
  return *this;
}

/*********************************************************************/
//the types of control locations
/*********************************************************************/
const int ContLoc::BREAK = 10000;
const int ContLoc::RETURN = 10010;
const int ContLoc::FINAL = 10020;
const int ContLoc::LABEL_ID = 10030;
const int ContLoc::LABEL_CASE = 10040;
const int ContLoc::LABEL_DEFAULT = 10050;
const int ContLoc::CONTINUE = 10060;
const int ContLoc::GOTO = 10070;
const int ContLoc::ASSIGN_EXPR = 10080;
const int ContLoc::ASSIGN_CALL = 10090;
const int ContLoc::BRANCH = 10110;

/*********************************************************************/
//constructors
/*********************************************************************/
ContLoc::ContLoc()
{
  locId = 0;
  indCallKey = -1;
  lhsAliasKey = -1;
  useful = false;
  procInfo = NULL;
  tSucc = eSucc = NULL;
  locType = 0;
  callSucc = NULL;
}

ContLoc::ContLoc(ProcInfo *p,const int t)
{
  locId = 0;
  indCallKey = -1;
  lhsAliasKey = -1;
  useful = false;
  procInfo = p;
  tSucc = eSucc = NULL;
  locType = t;
  callSucc = NULL;
}

ContLoc::ContLoc(const ContLoc &rhs)
{
  *this = rhs;
}

ContLoc::~ContLoc()
{
  Cleanup();
}

const ContLoc &ContLoc::operator = (const ContLoc &rhs)
{
  Cleanup();
  locId = rhs.locId;
  indCallKey = rhs.indCallKey;
  lhsAliasKey = rhs.lhsAliasKey;
  preds = rhs.preds;
  useful = rhs.useful;
  procInfo = rhs.procInfo;
  absProc = rhs.absProc;
  tSucc = rhs.tSucc;
  eSucc = rhs.eSucc;
  tSkipped = rhs.tSkipped;
  eSkipped = rhs.eSkipped;
  locType = rhs.locType;
  locExpr = rhs.locExpr;
  locLhs = rhs.locLhs;
  locRhs = rhs.locRhs;
  lhsAliases = rhs.lhsAliases;
  label = rhs.label;
  callSucc = rhs.callSucc;
  callSkipped = rhs.callSkipped;
  predecessors = rhs.predecessors;
  implStates = rhs.implStates;
  keyToImpl = rhs.keyToImpl;
  propMap = rhs.propMap;
  succsAndActions = rhs.succsAndActions;
  predsAndActions = rhs.predsAndActions;
  liveVars = rhs.liveVars;
  killSet = rhs.killSet;
  genSet = rhs.genSet;
  backup = rhs.backup;
  return *this;
}

void ContLoc::Cleanup()
{
  for(set<const ImplState*>::const_iterator i = implStates.begin();i != implStates.end();i++) {
    delete *i;
  }
  implStates.clear();
}

/*********************************************************************/
//get string representation
/*********************************************************************/
string ContLoc::ToString() const
{
  string a;
  
  if(locType == BREAK) a = "break location";
  else if(locType == RETURN) {
    a = "return ( " + locExpr.ToString() + ")";
  }
  else if(locType == FINAL) a = "final location";
  else if(locType == LABEL_ID) a = "label location";
  else if(locType == LABEL_CASE) a = "case location";
  else if(locType == LABEL_DEFAULT) a = "default location";
  else if(locType == CONTINUE) a = "continue location";
  else if(locType == GOTO) a = "goto location";
  else if((locType == ASSIGN_EXPR) || (locType == ASSIGN_CALL)) {
    assert(!locLhs.IsNull() && !locRhs.IsNull());
    a = locLhs.ToString() + "= " + Util::TrimString(locRhs.ToString());
  } else if(locType == BRANCH) {
    assert(!locExpr.IsNull());
    a = "branch ( " + locExpr.ToString() + ")";
  } else {
    assert(false);
    a = "unknown location";
  }

  return a;
}

/*********************************************************************/
//accessors and modifiers
/*********************************************************************/
void ContLoc::SetLocExpr(const Expr &e) { locExpr = e; }

void ContLoc::SetLocLhs(const Expr &e) { locLhs = e; }

void ContLoc::SetLocRhs(const Expr &e) { locRhs = e; }

void ContLoc::SetLhsAliases(const set<Expr> &s) { lhsAliases = s; }

/*********************************************************************/
//if this control location is a call-site then return true and also
//the components of the call i.e. the procedure name and the arguments
//to the called procedure in the array supplied as argument. otherwise
//return false.
/*********************************************************************/
bool ContLoc::IsCallSite(pair< Expr,list<Expr> > &comps) const
{
  if(locType == ASSIGN_CALL) {
    comps = ExprManager::GetCalledProcDetails(locRhs);
    return true;
  } else {
    return false;
  }
}

/*********************************************************************/
//returns true iff this is a skip location - break, label, case,
//default, continue, goto or a call location abstracted only by
//useless spec LTSs.
/*********************************************************************/
bool ContLoc::IsSkipLoc() const
{
  if ((locType == BREAK) || (locType == LABEL_ID) ||
      (locType == LABEL_CASE) || (locType == LABEL_DEFAULT) ||
      (locType == CONTINUE) || (locType == GOTO)) {
    return true;
  } else if(locType == ASSIGN_CALL) {
    //get the called procedure expression and name
    pair< Expr,list<Expr> > comps = ExprManager::GetCalledProcDetails(locRhs);
    //if this is an abstracted call-site
    pair< set<string>,set<string> > csi = procInfo->GetCallSiteInfo(comps.first,indCallKey);
    if(csi.first.empty()) {
      //get the set of names of LTSs that abstract this call-site
      set<string> snames;
      for(set<string>::const_iterator i = csi.second.begin();i != csi.second.end();i++) {
	if(Database::procInfos.count(*i) != 0) {
	  const list<ProcAbs> &pabs = Database::procInfos[*i].GetCallerAbs();
	  for(list<ProcAbs>::const_iterator j = pabs.begin();j != pabs.end();j++) {
	    snames.insert(j->GetLtsName());
	  }
	}
      }
      //check if any of the specs are useful - this location is skip
      //iff every spec LTS is useless
      for(set<string>::iterator i = snames.begin();i != snames.end();i++) {
	const LtsInfo &b = Database::GetLtsInfo(*i);
	if(b.IsUseful()) return false;
      }
      return true;
    }
    //otherwise this location is not skip
    else return false;
  } else return false;
}

/*********************************************************************/
//initialise the live variable analysis at this location
/*********************************************************************/
void ContLoc::InitLiveVarAnalysis()
{
  //initialise the live variable set to an empty set
  liveVars.clear();
  
  //create the kill set
  killSet.clear();
  if((locType == ASSIGN_EXPR) || (locType == ASSIGN_CALL)) {
    killSet.insert(Util::TrimString(locLhs.ToString()));
  }
  
  if(locType == RETURN) {
    CreateGenSetReturn();
  } else if(locType == ASSIGN_EXPR) {
    CreateGenSetAssignExpr();
  } else if(locType == ASSIGN_CALL) {
    CreateGenSetAssignCall();
  } else if(locType == BRANCH) {
    CreateGenSetBranch();
  } else genSet.clear();

  //for ltl model checking add the variables appearing in propositions
  if(Database::CONF_TYPE == Database::CONF_LTL) {
    const set<Expr> &ltlProps = procInfo->GetProps();
    for(set<Expr>::const_iterator i = ltlProps.begin();i != ltlProps.end();i++) {
      set<string> x = Util::ComputeLvalues(i->GetExpr());
      genSet.insert(x.begin(),x.end());
    }
  }
}

/*********************************************************************/
//create the genset for a retun control location
/*********************************************************************/
void ContLoc::CreateGenSetReturn()
{
  genSet.clear();
  genSet = Util::ComputeLvalues(locExpr.GetExpr());
}

/*********************************************************************/
//create the genset for an assignment location where the rhs is not a
//procedure call
/*********************************************************************/
void ContLoc::CreateGenSetAssignExpr()
{
  genSet = Util::ComputeLvalues(locRhs.GetExpr());
}

/*********************************************************************/
//create the genset for an assignment location where the rhs is a
//procedure call
/*********************************************************************/
void ContLoc::CreateGenSetAssignCall()
{
  genSet.clear();
  //get the call details
  pair< Expr,list<Expr> > comps = ExprManager::GetCalledProcDetails(locRhs);
  //add the variables in the arguments and the procedure name in the
  //genset
  list<Expr> &args = comps.second;
  for(list<Expr>::const_iterator i = args.begin();i != args.end();i++) {
    set<string> x = Util::ComputeLvalues(i->GetExpr());
    genSet.insert(x.begin(),x.end());
  }
  set<string> x = Util::ComputeLvalues(comps.first.GetExpr());
  genSet.insert(x.begin(),x.end());
  //add the variables in the guard predicates
  set<Expr> actPreds = procInfo->GetCallActPreds(comps.first,absProc);
  for(set<Expr>::const_iterator i = actPreds.begin();i != actPreds.end();i++) {
    //replace "$1", "$2" etc. with the actual arguments
    set<string> x = Util::ComputeLvalues(ExprManager::ReplaceDummyVars(*i,args,1).GetExpr());
    genSet.insert(x.begin(),x.end());
  }
}

/*********************************************************************/
//create the genset for a branch location
/*********************************************************************/
void ContLoc::CreateGenSetBranch()
{
  genSet = Util::ComputeIds(locExpr.GetExpr());
}

/*********************************************************************/
//make this control location a predecessor of each of its successor
/*********************************************************************/
void ContLoc::MakePred()
{
  //get the set of successor locations
  list<ContLoc*> x = ComputeSuccLocs();
  
  //make this location a predecessor of each successor
  for(list<ContLoc*>::iterator i = x.begin();i != x.end();i++) {
    (*i)->predecessors.insert(this);
  }
}

/*********************************************************************/
//compute the set of successor locations appropriate for live variable
//analysis
/*********************************************************************/
list<ContLoc*> ContLoc::ComputeSuccLocs()
{
  list<ContLoc*> x;
  if((locType == RETURN) || (locType == FINAL)) {}
  else if((locType == ASSIGN_EXPR) || (locType == ASSIGN_CALL)) {
    x.push_back(tSucc);
  } else if(locType == BRANCH) {
    x.push_back(tSucc);
    x.push_back(eSucc);
  } else assert(false);
  return x;
}

/*********************************************************************/
//update the set of live variables at this location. return true if
//the set is changed by the update and false otherwise.
/*********************************************************************/
bool ContLoc::UpdateLiveVars()
{
  //get the set of successor locations
  list<ContLoc*> x = ComputeSuccLocs();
  
  //collect the live variables at the successors
  set<string> y;
  for(list<ContLoc*>::iterator i = x.begin();i != x.end();i++) {
    set<string> &a = (*i)->liveVars;
    y.insert(a.begin(),a.end());
  }

  //remove the kill set and add the gen set
  for(set<string>::iterator i = killSet.begin();i != killSet.end();i++) {
    y.erase(*i);
  }
  y.insert(genSet.begin(),genSet.end());

  //compute if the set of live variables will be modified
  bool res = false;
  for(set<string>::iterator i = y.begin();i != y.end();i++) {
    if(liveVars.count(*i) == 0) {
      res = true;
      break;
    }
  }
  
  //update the set of live variables
  liveVars = y;
  
  return res;
}

/*********************************************************************/
//add the set of predicates associated with return actions. this set
//of predicates is supplied as the first argument.
/*********************************************************************/
void ContLoc::AddRetActPreds(const set<Expr> &actPreds)
{
  //add the predicates after replacing all occurrences of "$0" with
  //the return expression and "$1", "$2" etc. with the parameters of
  //the procedure
  for(set<Expr>::const_iterator i = actPreds.begin();i != actPreds.end();i++) {
    list<Expr> repList = ExprManager::StringListToExprList(procInfo->GetParams());
    repList.push_front(locExpr);
    Expr res = ExprManager::ReplaceDummyVars(*i,repList,0);

    //create the predicate and add it to this location
    if(AddPred(Predicate(res,Predicate::REQUIRED,pair<const ContLoc*,int>(this,1)))) {
      Util::Message(3,"predicate " + res.ToString() + " added to control location " + ToString() + "\n");
    }
  }
}

/*********************************************************************/
//add a predicate to this control location. return true if the set of
//predicates was updated and false otherwise.
/*********************************************************************/
bool ContLoc::AddPred(const Predicate &p)
{
  //sanitize the predicate
  Predicate pred = p.Sanitize(this);

  //first check if the predicate has any live variables as a lvalue
  if(pred.ContainsLvalue(liveVars)) {
    return preds.AddPred(pred);
  } else {
    return false;
  }
}

/*********************************************************************/
//add the set of predicates associated with call actions. the argument
//is a map from procedure names to sets of predicates associated with
//calls to those procedures
/*********************************************************************/
void ContLoc::AddCallActPreds()
{
  //get the components to the call i.e. the procedure name and the arguments
  pair< Expr,list<Expr> > comps;
  if(locType == ASSIGN_CALL) comps = ExprManager::GetCalledProcDetails(locRhs);
  else Util::Error("trying to get components of non-procedure call : " + ToString() + "\n");  
  //get the set of predicates associated with this call
  set<Expr> actPreds =  procInfo->GetCallActPreds(comps.first,absProc);
  //do the replacement and add the predicate
  list<Expr> &args = comps.second;
  for(set<Expr>::const_iterator i = actPreds.begin();i != actPreds.end();i++) {
    //replace "$1", "$2" etc. with the actual arguments
    Expr b = ExprManager::ReplaceDummyVars(*i,args,1);
    if(AddPred(Predicate(b,Predicate::REQUIRED,pair<const ContLoc*,int>(this,1)))) {
      Util::Message(3,"predicate " + b.ToString() + " added to control location " + ToString() + "\n");
    } 
  }
}

/*********************************************************************/
//update the predicates at this control location. return true if the
//predicates were updated and false otherwise.
/*********************************************************************/
bool ContLoc::UpdatePreds()
{
  //return location
  if(locType == RETURN) return UpdatePredsReturn();
  //final location
  else if(locType == FINAL) return UpdatePredsFinal();
  //assign with rhs being not a procedure call
  else if(locType == ASSIGN_EXPR) return UpdatePredsAssignExpr();
  //assign with rhs being a procedure call
  else if(locType == ASSIGN_CALL) return UpdatePredsAssignCall();
  //branch statement
  else if(locType == BRANCH) return UpdatePredsBranch();
  //unknown location
  else assert(false);
  return false;
}

/*********************************************************************/
//no update required at return control locations. the seeding
//predicates are enough.
/*********************************************************************/
bool ContLoc::UpdatePredsReturn()
{
  return false;
}

/*********************************************************************/
//update the predicates at a final control location. return true if
//the predicates were updated and false otherwise.
/*********************************************************************/
bool ContLoc::UpdatePredsFinal()
{
  //the final location does not need any predicates
  return false;
}

/*********************************************************************/
//update the predicates at a assign control location with the rhs
//being not a procedure call. return true if the predicates were
//updated and false otherwise.
/*********************************************************************/
bool ContLoc::UpdatePredsAssignExpr()
{
  //add the weakest preconditions of the predicates at the
  //successor location under the assignment
  list<Expr> lhsList,rhsList;
  lhsList.push_back(locLhs);
  rhsList.push_back(locRhs);

  const vector<Predicate> &succPreds = tSucc->preds.GetPreds();
  bool flag = false;
  for(vector<Predicate>::const_iterator i = succPreds.begin();i != succPreds.end();i++) {
    const Predicate &succPred = *i;
    //create the weakest precondition. if the predicate is
    //being inferred from a predicate of non-REQUIRED type
    //then add one to the type.
    int newType = (succPred.GetType() == Predicate::REQUIRED) ? Predicate::REQUIRED : (succPred.GetType() + 1);
    Predicate predWP = succPred.ComputeWP(lhsList,rhsList);
    predWP.SetType(newType);
    //if the computed WP is different then this location is useful
    if(!(predWP == succPred)) useful = true;
    if(AddPred(predWP)) {
      Util::Message(3,"predicate " + predWP.ToString() + " added to control location " + ToString() + "\n");
      flag = true;
    }
  }
  return flag;
}

/*********************************************************************/
//update the predicates at a assign control location with the rhs
//being a procedure call. return true if the predicates were updated
//and false otherwise.
/*********************************************************************/
bool ContLoc::UpdatePredsAssignCall()
{
  //we just inherit the predicates from the successor 
  const vector<Predicate> &succPreds = tSucc->preds.GetPreds();
  bool flag = false;
  for(vector<Predicate>::const_iterator i = succPreds.begin();i != succPreds.end();i++) {
    if(AddPred(*i)) {
      Util::Message(3,"predicate " + i->ToString() + " added to control location " + ToString() + "\n");
      flag = true;
    }
  }
  return flag;
}

/*********************************************************************/
//update the predicates at a branch control location. return true if
//the predicates were updated and false otherwise.
/*********************************************************************/
bool ContLoc::UpdatePredsBranch()
{
  bool flag = false;
  
  //if no predicates have been added just add the branch
  //condition
  if((!(Database::CEGAR && Database::USE_PRED_ABSTRACTION)) && preds.GetPreds().empty()) {
    //if the spec files contain the initial predicates then
    //check if this branch condition is one of them or the
    //negation of one of them. add the branch condition to
    //this location only if that condition is satisfied.
    Predicate condPred(locExpr,Predicate::TENTATIVE,pair<const ContLoc*,int>(this,1));	    
    if(Database::PREDS_FROM_SPEC) {
      if(useful) {
	if(AddPred(condPred)) {
	  Util::Message(3,"predicate " + condPred.ToString() + " added to control location " + ToString() + "\n");
	  flag = true;
	}
      }
    } else {
      if(AddPred(condPred)) {
	Util::Message(3,"predicate " + condPred.ToString() + " added to control location " + ToString() + "\n");
	flag = true;
      }
    }
  }
  
  if(MergePredsConjunct(tSucc,locExpr)) flag = true;
  Expr nexpr = ExprManager::NegateExpr(locExpr);
  if(MergePredsConjunct(eSucc,nexpr)) flag = true;
  return flag;
}

/*********************************************************************/
//add the conjunction of the predicates of a successor location with
//the given expression.
/*********************************************************************/
bool ContLoc::MergePredsConjunct(ContLoc *loc,const Expr &expr)
{
  const vector<Predicate> &a = loc->preds.GetPreds();
  bool flag = false;
  for(vector<Predicate>::const_iterator i = a.begin();i != a.end();i++) {
    Predicate c = i->Conjunct(expr);
    if((c.GetSize() != 0) && AddPred(c)) {
      Util::Message(3,"predicate " + c.ToString() + " added to control location " + ToString() + "\n");
      flag = true;
    }
  }
  return flag;
}

/*********************************************************************/
//compute if this location is useful or not. this is only applicable
//for ASSIGN_EXPR locations. the useful bit has been partially
//computed during the previous relevant predicate computation
//phase. now we make certain useless locations useful if they can
//cause assign actions.
/*********************************************************************/
void ContLoc::ComputeUseful()
{
  if(locType != ASSIGN_EXPR) return;
  else if(Database::NO_USELESS_LOCS) {
    for(PredVal i = preds.GetInitPredValuation();preds.IsValidPredValuation(i);i = preds.GetNextPredValuation(i)) {
      set<Expr> state = preds.ToExpr(i);
      set<Action> acts = Database::GetSpecAssignActions(locLhs,locRhs,procInfo->GetParams(),state);
      if(!acts.empty()) {
	useful = true;
	return;
      }
    }
  } else {
    useful = true;
  }
}

/*********************************************************************/
//if the first argument is a successor of this locations then replace
//it with the second argument.
/*********************************************************************/
void ContLoc::ReplaceSucc(ContLoc *prev,ContLoc *next)
{
  assert(prev != next);
  if((locType == RETURN) || (locType == FINAL)) {}
  else if(locType == ASSIGN_EXPR) {
    if(tSucc == prev) {
      tSkipped.push_back(tSucc);
      tSucc = next;
    }
  } else if(locType == ASSIGN_CALL) {
    if(tSucc == prev) {
      tSkipped.push_back(tSucc);
      tSucc = next;
    }
    if(callSucc == prev) {
      callSkipped.push_back(callSucc);
      callSucc = next;
    }
  } else if(locType == BRANCH) {
    if(tSucc == prev) {
      tSkipped.push_back(tSucc);
      tSucc = next;
    }
    if(eSucc == prev) {
      eSkipped.push_back(eSucc);
      eSucc = next;
    }
  } else {
    assert(false);
  }
}

/*********************************************************************/
//returns true if this is an abstracted call site
/*********************************************************************/
bool ContLoc::IsAbsCallSite() const
{
  pair< Expr,list<Expr> > comps;
  if(IsCallSite(comps)) {
    assert(!procInfo->GetCallSiteInfo(comps.first,indCallKey).second.empty());
    return true;
  } else return false;
}

/*********************************************************************/
//compute the spec infos corresponding to this control location
/*********************************************************************/
void ContLoc::ComputeSpecInfos()
{
  //if this is an abstracted call-site then each implementation
  //state will also have a LTS state component
  if(IsAbsCallSite()) {
    set<string> allSpecs;
    for(PredVal i = preds.GetInitPredValuation();preds.IsValidPredValuation(i);
	i = preds.GetNextPredValuation(i)) {
      //get the names of the spec LTSs that abstract this call
      set<Expr> a = preds.ToExpr(i);
      set<string> ltsNames = procInfo->GetCallActions(locRhs,absProc,a);
      allSpecs.insert(ltsNames.begin(),ltsNames.end());
    }
    //create the lts infos
    for(set<string>::iterator it = allSpecs.begin();it != allSpecs.end();it++) {
      Database::GetLtsInfo(*it);
    }
  }
}

/*********************************************************************/
//compute succs and actions of reachable impl states
/*********************************************************************/
void ContLoc::ComputeSuccsAndActions()
{
  for(set<const ImplState*>::const_iterator i = implStates.begin();i != implStates.end();i++) {
    (*i)->GetSuccsAndActions();
  }
}

/*********************************************************************/
//restore the successors and actions of this control location from the
//file whose pointer is passed as argument
/*********************************************************************/
void ContLoc::RestoreSuccsAndActions(FILE *in)
{
  while(true) {
    ContLoc *newLoc = NULL;
    unsigned long oldPVL,newPVL;
    short oldSS,oldSI,newSS,newSI,actId;
    fscanf(in,"%lu",&oldPVL);
    fscanf(in,"%hd",&oldSS);
    fscanf(in,"%hd",&oldSI);
    fscanf(in,"%hd",&actId);
    fscanf(in,"%p",&newLoc);
    fscanf(in,"%lu",&newPVL);
    fscanf(in,"%hd",&newSS);
    fscanf(in,"%hd",&newSI);
    unsigned long dummy = PredSet::PValToUL(PVal().flip());
    if((oldPVL == dummy) && (oldSS == -1) && (oldSI == -1) && (actId == Action::ACTION_NONE) && 
       (newLoc == 0x0) && (newPVL == dummy) && (newSS == -1) && (newSI == -1)) break;
    PVal oldPV = PredSet::ULToPVal(oldPVL);
    PVal newPV = PredSet::ULToPVal(newPVL);
    const ImplState *oldState = GetImplState(oldPV,oldSS,oldSI);
    string oldKey = oldState->GetKey();
    const ImplState *newState = newLoc->GetImplState(newPV,newSS,newSI);
    succsAndActions[oldKey].push_back(pair<const ImplState*,Action>(newState,Action(actId)));
    string newKey = newState->GetKey();
    newLoc->predsAndActions[newKey].push_back(pair<const ImplState*,Action>(oldState,Action(actId)));
  }
}

/*********************************************************************/
//backup the successors and actions of this control location to the
//file whose pointer is passed as argument
/*********************************************************************/
void ContLoc::BackupSuccsAndActions(FILE *out)
{
  for(set<const ImplState*>::const_iterator i = implStates.begin();i != implStates.end();i++) {
    unsigned long oldPVL = PredSet::PValToUL((*i)->GetPredVal());
    short oldSS = (*i)->GetSpecStateId();
    short oldSI = (*i)->GetSpecInfoId();
    list< pair<const ImplState*,Action> > succs = (*i)->GetSuccsAndActions();
    for(list< pair<const ImplState*,Action> >::const_iterator j = succs.begin();j != succs.end();j++) {
      ContLoc *newLoc = j->first->GetContLoc();
      unsigned long newPVL = PredSet::PValToUL(j->first->GetPredVal());
      short newSS = j->first->GetSpecStateId();
      short newSI = j->first->GetSpecInfoId();
      short actId = j->second.GetIndex();
      fprintf(out,"%lu %hd %hd %hd %p %lu %hd %hd ",oldPVL,oldSS,oldSI,actId,newLoc,newPVL,newSS,newSI);
    }  
  }
  unsigned long dummy = PredSet::PValToUL(PVal().flip());
  fprintf(out,"%lu -1 -1 %hd 0x0 %lu -1 -1 ",dummy,Action::ACTION_NONE,dummy);
}

/*********************************************************************/
//returns true iff the succs and actions map contains an entry for
//this key
/*********************************************************************/
bool ContLoc::ContainsSuccsAndActions(const string &key)
{
  return (succsAndActions.count(key) != 0);
}

/*********************************************************************/
//set the propositions for a given impl state
/*********************************************************************/
void ContLoc::SetPropositions(const ImplState *is,const set<Expr> &props)
{
  assert(propMap.count(is) == 0);
  propMap[is] = props;
}

/*********************************************************************/
//return the set of propositions true in the implementation state with
//given key
/*********************************************************************/
set<Expr> ContLoc::GetPropositions(const string &key)
{
  map<string,const ImplState*>::const_iterator i = keyToImpl.find(key);
  assert(i != keyToImpl.end());
  map< const ImplState*,set<Expr> >::const_iterator j = propMap.find(i->second);
  assert(j != propMap.end());
  return j->second;
}

/*********************************************************************/
//returns the entry for the key in the succs and actions map
/*********************************************************************/
const list< pair<const ImplState*,Action> > &ContLoc::GetSuccsAndActions(const string &key)
{
  assert(succsAndActions.count(key) != 0);
  return succsAndActions[key];
}

/*********************************************************************/
//update the successors and actions map
/*********************************************************************/
void ContLoc::AddSuccsAndActions(const string &key,const list< pair<const ImplState*,Action> > &val)
{
  assert(succsAndActions.count(key) == 0);
  succsAndActions[key] = val;
  map<string,const ImplState*>::const_iterator i = keyToImpl.find(key);
  assert(i != keyToImpl.end());
  for(list< pair<const ImplState*,Action> >::const_iterator j = val.begin();j != val.end();j++) {
    ContLoc *newLoc = j->first->GetContLoc();
    string newKey = j->first->GetKey();
    newLoc->predsAndActions[newKey].push_back(pair<const ImplState*,Action>(i->second,j->second));
  }
}

/*********************************************************************/
//returns the entry for the key in the preds and actions map
/*********************************************************************/
list< pair<const ImplState*,Action> > ContLoc::GetPredsAndActions(const string &key)
{
  if(predsAndActions.count(key) == 0) return list< pair<const ImplState*,Action> >();
  else return predsAndActions[key];
}

/*********************************************************************/
//return the implementation states given a predicate valuation
/*********************************************************************/
set<const ImplState*> ContLoc::GetImplStates(const PVal &pv)
{
  set<const ImplState*> results;

  //if this is a call site
  if(IsAbsCallSite()) {
    set<Expr> conc = preds.ToExpr(preds.PValToPredVal(pv));
    //get the names of the spec LTSs that abstract this call
    set<string> ltsNames = procInfo->GetCallActions(locRhs,absProc,conc);
    for(set<string>::const_iterator j = ltsNames.begin();j != ltsNames.end();j++) {
      const LtsInfo &li = Database::GetLtsInfo(*j);
      string key = ImplState(locId,pv,li.GetStateId(*j),li.GetIndex()).GetKey();
      if(keyToImpl.count(key) == 0) {
	const ImplState *is = new ImplState(locId,pv,li.GetStateId(*j),li.GetIndex());
	implStates.insert(is);
	keyToImpl[key] = is;
	results.insert(is);
      } else {
	results.insert(keyToImpl[key]);
      }
    }
  }
  //otherwise
  else {
    string key = ImplState(locId,pv,-1,-1).GetKey();
    if(keyToImpl.count(key) == 0) {
      const ImplState *is = new ImplState(locId,pv,-1,-1);
      implStates.insert(is);
      keyToImpl[key] = is;
      results.insert(is);
    } else {
      results.insert(keyToImpl[key]);
    }
  }
  //all done
  return results;
}

/*********************************************************************/
//returns the initial implementation states matching a predicate
/*********************************************************************/
set<const ImplState*> ContLoc::GetImplStates(const Expr &pred)
{
  set<const ImplState*> results;
  assert(!pred.IsNull());

  //get the ast for the negation of the predicate
  Expr neg = ExprManager::NegateExpr(pred);
  set<Expr> eset; eset.insert(neg);

  //if this is a call site
  if(IsAbsCallSite()) {
    for(PredVal i = preds.GetInitPredValuation();preds.IsValidPredValuation(i);
	i = preds.GetNextPredValuation(i)) {
      set<Expr> conc = preds.ToExpr(i);
      if(!ExprManager::ProveImplies(conc,eset)) {
	//get the names of the spec LTSs that abstract this call
	set<string> ltsNames = procInfo->GetCallActions(locRhs,absProc,conc);
	for(set<string>::const_iterator j = ltsNames.begin();j != ltsNames.end();j++) {
	  const LtsInfo &li = Database::GetLtsInfo(*j);
	  string key = ImplState(locId,i.first,li.GetStateId(*j),li.GetIndex()).GetKey();
	  if(keyToImpl.count(key) == 0) {
	    const ImplState *is = new ImplState(locId,i.first,li.GetStateId(*j),li.GetIndex());
	    implStates.insert(is);
	    keyToImpl[key] = is;
	    results.insert(is);
	  } else {
	    results.insert(keyToImpl[key]);
	  }
	}
      }
    }    
  }
  //otherwise
  else {
    for(PredVal i = preds.GetInitPredValuation();preds.IsValidPredValuation(i);
	i = preds.GetNextPredValuation(i)) {
      set<Expr> conc = preds.ToExpr(i);
      if(!ExprManager::ProveImplies(conc,eset)) {
	string key = ImplState(locId,i.first,-1,-1).GetKey();
	if(keyToImpl.count(key) == 0) {
	  const ImplState *is = new ImplState(locId,i.first,-1,-1);
	  implStates.insert(is);
	  keyToImpl[key] = is;
	  results.insert(is);
	} else {
	  results.insert(keyToImpl[key]);
	}
      }
    }
  }
  //all done
  return results;
}

/*********************************************************************/
//return the implementation state given a predicate valuation and a
//LTS state id.
/*********************************************************************/
const ImplState *ContLoc::GetImplState(const PVal &pv,short st,short sp)
{
  string key = ImplState(locId,pv,st,sp).GetKey();
  if(keyToImpl.count(key) == 0) {
    const ImplState *is = new ImplState(locId,pv,st,sp);
    implStates.insert(is);
    keyToImpl[key] = is;
    return is;
  } else {
    return keyToImpl[key];
  }
}

/*********************************************************************/
//backup important fields
/*********************************************************************/
void ContLoc::Backup()
{
  backup.tSucc = tSucc;
  backup.eSucc = eSucc;
  backup.callSucc = callSucc;
}

/*********************************************************************/ 
//restore all important fields of the control locations,cleanup
//implementation states, specification state names, successor
//implementation states and simulation relation from the previous
//iteration and set the useful fields to false
/*********************************************************************/
void ContLoc::Restore()
{
  //restore fields
  tSucc = backup.tSucc;
  eSucc = backup.eSucc;
  callSucc = backup.callSucc;
  for(set<const ImplState*>::const_iterator i = implStates.begin();i != implStates.end();i++) {
    delete *i;
  }
  implStates.clear();
  keyToImpl.clear();
  propMap.clear();
  succsAndActions.clear();
  predsAndActions.clear();

  //sanity check
  assert(implStates.empty() && keyToImpl.empty() && succsAndActions.empty() && predsAndActions.empty());
}

/*********************************************************************/
//end of ContLoc.cpp
/*********************************************************************/
