/******************************** CPPFile *****************************

* FileName [Database.cpp]

* PackageName [main]

* Synopsis [This file contains procedures to access and manipulate
* global data structures.]

* SeeAlso [Database.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <sys/time.h>
#include <string>
#include <list>
#include <map>
#include <set>
#include <vector>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Statistics.h"
#include "CSymbolTable.h"
#include "Node.h"
#include "DFSAdapter.h"
#include "Action.h"
#include "Alias.h"
#include "ProgAbs.h"
#include "ProgInfo.h"
#include "ProcAbs.h"
#include "ProcAddInfo.h"
#include "LtsTrans.h"
#include "FspInfo.h"
#include "LtlFormula.h"
#include "LtlManager.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "ProcInfo.h"
#include "LtsInfo.h"
#include "CallGraph.h"
#include "CallGraphCreator.h"
#include "AbsRefiner.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "Buchi.h"
#ifndef MAGIC_LITE
#include "ctool.h"
#include "CtoolAPI.h"
#include "PamAPI.h"
using namespace ctool;
#endif //MAGIC_LITE

/*********************************************************************/
//define static constants
/*********************************************************************/
const string Database::MAGIC_VERSION = "0.1";
const short int Database::PRE_INC = 6000;
const short int Database::PRE_DEC = 6010;
const short int Database::POST_INC = 6020;
const short int Database::POST_DEC = 6030;
const string Database::EPSILON = "epsilon";
const string Database::EPSILON_SPEC = "EpsilonSpec";
const string Database::EPSILON_PROC_ABS = "EpsilonProcAbs";
const string Database::DEFAULT_SPEC = "DefaultSpec";
const string Database::STOP_STATE = "STOP";
const string Database::ERROR_STATE = "ERROR";
const string Database::DUMMY_C_PROC = "MagicDummyCProc";
const int Database::CONF_SIMUL = 2500;
const int Database::CONF_REACH = 2510;
const int Database::CONF_TRACE = 2520;
const int Database::CONF_LTL = 2530;
const int Database::CONF_CHECK_EXPLICIT = 1000;
const int Database::CONF_CHECK_SAT = 1010;
const int Database::C_PARSER_DEFAULT = 8000;
const int Database::C_PARSER_CTOOL = 8010;
const int Database::THEOREM_PROVER_SIMPLIFY = 5000;
const int Database::THEOREM_PROVER_CPROVER = 5010;
const int Database::THEOREM_PROVER_CVC = 5020;
const int Database::THEOREM_PROVER_ICS = 5030;
const int Database::THEOREM_PROVER_SVC = 5040;
const int Database::SAT_SOLVER_CHAFF = 5500;
const int Database::SAT_SOLVER_SATO = 5510;
const int Database::SAT_SOLVER_GRASP = 5520;
const int Database::SAT_SOLVER_HORN_AI = 5530;
const int Database::PATH_CHECK_FORWARD = 5700;
const int Database::PATH_CHECK_BACKWARD = 5710;
const int Database::PATTERN_MATCHER_NAIVE = 7000;
const int Database::CEGAR_ELIMINATE_PRED = 7100;
const int Database::CEGAR_USEFUL_PRED = 7110;
const int Database::CEGAR_OPTIMIZE_PRED = 7120;
const int Database::CEGAR_GREEDY_PRED = 7130;
const size_t Database::PVBITS;
//const int Database::COVER_NUMBER_TRACES = 1000;
//const int Database::NUMBER_TRACES = 1000;
//const int Database::TRACES_LENGTH = 500;
//const Random Database::rnd = new Random ();
const string Database::DOT_FILE_NAME = "magic.dot";
const int Database::OUTPUT_MODEL_FSP = 9500;
const int Database::REFINE_NONE = 9900;
const int Database::REFINE_PRED = 9910;
const int Database::REFINE_LTS = 9920;

/*********************************************************************/
//define static arguments and options
/*********************************************************************/
list<string> Database::FILE_NAMES;
string Database::ABSTRACTION_NAME;
string Database::SIM_REL_FILE;
bool Database::PARSE_ONLY = false;
bool Database::PARSE_ECHO = false;
bool Database::DISPLAY_STATISTICS = false;
int Database::CONF_TYPE = CONF_SIMUL;
int Database::CONF_CHECK_MECHANISM = CONF_CHECK_SAT;
int Database::C_PARSER_TYPE = C_PARSER_DEFAULT;
int Database::THEOREM_PROVER_USED = THEOREM_PROVER_SIMPLIFY;
bool Database::USE_PAM = false;
int Database::SAT_SOLVER_USED = SAT_SOLVER_HORN_AI;
int Database::PATTERN_MATCHER_USED = PATTERN_MATCHER_NAIVE;
bool Database::USE_SYNTACTIC_CHECKS = true;
bool Database::THEOREM_PROVER_CACHE = false;
int Database::THEOREM_PROVER_CACHE_SIZE = 100000;
bool Database::PREDS_FROM_SPEC = true;
int Database::VERBOSITY_LEVEL = 2;
bool Database::USE_REACH_IMPL = true;
bool Database::PRECOMPUTE_IMPL_TRANS = true;
bool Database::CACHE_IMPL_TRANS = true;
bool Database::SIM_REL_FROM_SAT = false;
bool Database::PREOPTIMIZE = false;
bool Database::PREOPTVISIT = false;
int Database::PREOPTMINDEPTH = 50;
size_t Database::PREOPTMINCES = 100;
int Database::PREOPTMAXDEPTH = 500;
size_t Database::PREOPTMAXCES = 2000;
int Database::PREOPTVISITS = 5;
bool Database::COVERAGE = false;
bool Database::COVERCONT = false;
bool Database::SHOWCOVER = false;
bool Database::CEGAR = false;
int Database::CEGAR_TYPE = CEGAR_ELIMINATE_PRED;
bool Database::INC_VERIFY = true;
int Database::MAX_ADD_PRED = 1;
bool Database::CHECK_CE_BRANCHING = false;
int Database::PATH_CHECK_TYPE = PATH_CHECK_FORWARD;
bool Database::USE_PATTERN_MATCHER = false;
size_t Database::MAX_GLOBAL_CE = 1;
size_t Database::MAX_CE_LOOP = 0;
double Database::CE_NEW_NODE_PROB = 0.0;
bool Database::CNF_NO_NEW_VARS = true;
size_t Database::LARGEST_SUBSET = 10;
size_t Database::SMALLEST_SUBSET = 1;
size_t Database::SUBSET_CUTOFF = 1000;
size_t Database::CONSTRAINT_CUTOFF = 20;
bool Database::STOP_OPT_AFTER_ELIM = true;
bool Database::USE_PRED_ABSTRACTION = true;
bool Database::USE_LTS_ABSTRACTION = false;
bool Database::NO_USELESS_LOCS = false;
int Database::MAX_PRED_INFER_LOOP = 1;
int Database::MAX_TIME_LIMIT = 36000000;
bool Database::LTL_TO_BUCHI = false;
bool Database::CREATE_DOT_FILE = false;
bool Database::DRAW_LTS = false;
bool Database::DRAW_CFG = false;
bool Database::DRAW_INFERRED_PREDS = false;
bool Database::DRAW_PRED_ABS_LTS = false;
bool Database::DRAW_LTS_ABS_LTS = false;
bool Database::DRAW_COMPOSED_LTS = false;
bool Database::DRAW_CE_DAG = false;
bool Database::DRAW_CE_PROJ_LTS = false;
bool Database::DRAW_CE_PROJ_PRED = false;
bool Database::OUTPUT_MODELS = false;
int Database::OUTPUT_MODEL_TYPE = OUTPUT_MODEL_FSP;

/*********************************************************************/
//define static data structures
/*********************************************************************/
list<string> Database::commandLine;
map<string,StdcFile*> Database::cAsts;
set<string> Database::specFiles;
map<string,ProcAddInfo> Database::procAddInfos;
map<string,ProgInfo> Database::progInfos;
FspInfo Database::fspInfo;
map<string,const BasicExpr*> Database::enums;
map<string,ProcInfo> Database::procInfos;
string Database::specLtsName;
string Database::progName;
vector<string> Database::rootProcNames;
vector<Expr> Database::absGuards;
vector<LtsInfo> Database::ltsInfos;
map<string,string> Database::ltsNegMap;
CallGraph Database::callGraph;
vector<ProcManager> Database::procManagers;
int Database::lastRefType = REFINE_NONE;
int Database::lastRefComp = -1;
double Database::currTime = 0;
map<Expr,int> Database::exprPriority;
string Database::drawFileName = "magic.ps";
FILE *Database::dotFilePtr = NULL;
string Database::modelFileName = "magic.fsp";
FILE *Database::modelFilePtr = NULL;

/*********************************************************************/
//global shutdown method defined in arc.cpp needed
/*********************************************************************/
__attribute__ ((noreturn)) void GlobalShutdown(int sig);

/*********************************************************************/
//parse command line and get user specified values of various options
/*********************************************************************/
void Database::ParseCommandLine()
{
  for(list<string>::const_iterator i = commandLine.begin();
      i != commandLine.end();i++) {
    if(((*i) == "--help") || ((*i) == "-h")) {
      DisplayUsage();
      GlobalShutdown(-1);
    } else if((*i) == "--parse") {
      PARSE_ONLY = true;
    } else if((*i) == "--echo") {
      PARSE_ECHO = true;
    } else if((*i) == "--stat") {
      DISPLAY_STATISTICS = true;
    } else if((*i) == "--simFile") {
      if((++i) != commandLine.end()) {
	SIM_REL_FILE = (*i);
      } else --i;
    } else if((*i) == "--reach") {
      CONF_TYPE = CONF_REACH;
    } else if((*i) == "--trace") {
      CONF_TYPE = CONF_TRACE;
    } else if((*i) == "--ltl") {
      CONF_TYPE = CONF_LTL;
    } else if((*i) == "--explicit") {
      CONF_CHECK_MECHANISM = CONF_CHECK_EXPLICIT;
    } else if((*i) == "--autoPred") {
      PREDS_FROM_SPEC = false;
    } else if((*i) == "--verbose") {
      if((++i) != commandLine.end()) {
	VERBOSITY_LEVEL = atoi(i->c_str());
	if(VERBOSITY_LEVEL < 0) {
	  Util::Error("verbosity level cannot be negative ...\n");
	}
      } else --i;
    } else if((*i) == "--noReachImpl") {
      USE_REACH_IMPL = false;
    } else if((*i) == "--abstraction") {
      if((++i) != commandLine.end()) {
	ABSTRACTION_NAME = (*i);
      } else --i;
    } else if((*i) == "--TPCache") {
      THEOREM_PROVER_CACHE = true;
    } else if((*i) == "--TPCacheSize") {
      if((++i) != commandLine.end()) {
	THEOREM_PROVER_CACHE_SIZE = atoi(i->c_str());
      } else --i;
    } else if((*i) == "--noImplTrans") {
      PRECOMPUTE_IMPL_TRANS = false;
    } else if((*i) == "--satSimRel") {
      SIM_REL_FROM_SAT = true;
    } else if((*i) == "--cprover") {
      THEOREM_PROVER_USED = THEOREM_PROVER_CPROVER;
    } else if((*i) == "--cvc") {
      THEOREM_PROVER_USED = THEOREM_PROVER_CVC;
    } else if((*i) == "--ics") {
      THEOREM_PROVER_USED = THEOREM_PROVER_ICS;
    } else if((*i) == "--svc") {
      THEOREM_PROVER_USED = THEOREM_PROVER_SVC;
    } else if((*i) == "--sato") {
      SAT_SOLVER_USED = SAT_SOLVER_SATO;
    } else if((*i) == "--grasp") {
      SAT_SOLVER_USED = SAT_SOLVER_GRASP;
    } else if((*i) == "--chaff") {
      SAT_SOLVER_USED = SAT_SOLVER_CHAFF;
    } else if((*i) == "--seed") {
      if((++i) != commandLine.end()) {
	//rnd.setSeed(Long.parseLong(*i));
      } else --i;
    } else if((*i) == "--preOpt") {
      PREOPTIMIZE = true;
    } else if((*i) == "--preOptVisit") {
      PREOPTVISIT = true;
    } else if((*i) == "--optPred") {
      CEGAR_TYPE = CEGAR_OPTIMIZE_PRED;
    } else if((*i) == "--usefulPred") {
      CEGAR_TYPE = CEGAR_USEFUL_PRED;
    } else if((*i) == "--greedy") {
      CEGAR_TYPE = CEGAR_GREEDY_PRED;
    } else if((*i) == "--largestSubset") {
      if((++i) != commandLine.end()) {
	LARGEST_SUBSET = atoi(i->c_str());
	if(LARGEST_SUBSET < 1) {
	  Util::Error("largest subset cannot be less than 1 ...\n");
	}
      } else --i;
    } else if((*i) == "--noInc") {
      INC_VERIFY = false;
    } else if((*i) == "--smallestSubset") {
      if((++i) != commandLine.end()) {
	SMALLEST_SUBSET = atoi(i->c_str());
	if(SMALLEST_SUBSET < 1) {
	  Util::Error("smallest subset cannot be less than 1 ...\n");
	}
      } else --i;
    } else if((*i) == "--subsetCutoff") {
      if((++i) != commandLine.end()) {
	SUBSET_CUTOFF = atoi(i->c_str());
	if(SUBSET_CUTOFF < 1) {
	  Util::Error("subset cutoff cannot be less than 1 ...\n");
	}
      } else --i;
    } else if((*i) == "--constraintCutoff") {
      if((++i) != commandLine.end()) {
	CONSTRAINT_CUTOFF = atoi(i->c_str());
	if(CONSTRAINT_CUTOFF < 1) {
	  Util::Error("constraint cutoff cannot be less than 1 ...\n");
	}
      } else --i;
    } else if((*i) == "--noStopAfterElim") {
      STOP_OPT_AFTER_ELIM = false;
    } else if((*i) == "--noNewVars") {
      CNF_NO_NEW_VARS = true;
    } else if((*i) == "--coverage") {
      COVERAGE = true;
    } else if((*i) == "--coverCont") {
      COVERCONT = true;
    } else if((*i) == "--showCoverage") {
      SHOWCOVER = true;
    } else if((*i) == "--cegar") {
      CEGAR = true;
    } else if((*i) == "--branching") {
      CHECK_CE_BRANCHING = true;
    } else if((*i) == "--pathBack") {
      PATH_CHECK_TYPE = PATH_CHECK_BACKWARD;
    } else if((*i) == "--patMatch") {
      USE_PATTERN_MATCHER = true;
    } else if((*i) == "--noPredAbs") {
      USE_PRED_ABSTRACTION = false;
    } else if((*i) == "--ltsAbs") {
      USE_LTS_ABSTRACTION = true;
    } else if((*i) == "--noUselessLocs") {
      NO_USELESS_LOCS = true;
    } else if((*i) == "--ceDag") {
      if((++i) != commandLine.end()) {
	MAX_GLOBAL_CE = atoi(i->c_str());
	if(MAX_GLOBAL_CE < 1) {
	  Util::Error("max global CE dag cannot be less than 1 ...\n");
	}
      } else --i;
    } else if((*i) == "--noTransCache") {
      CACHE_IMPL_TRANS = false;
    } else if((*i) == "--predLoop") {
      if((++i) != commandLine.end()) {
	MAX_PRED_INFER_LOOP = atoi(i->c_str());
	if(MAX_PRED_INFER_LOOP < 1) {
	  Util::Error("predicate inference loop cannot be less than 1 ...\n");
	}
      } else --i;
    } else if((*i) == "--noSyntactic") {
      USE_SYNTACTIC_CHECKS = false;
    } else if((*i) == "--ctool") {
      C_PARSER_TYPE = C_PARSER_CTOOL;
    } else if((*i) == "--pam") {
      USE_PAM = true;
    } else if((*i) == "--ltlToBuchi") {
      LTL_TO_BUCHI = true;
    } else if((*i) == "--drawLTS") {
      DRAW_LTS = true;
    } else if((*i) == "--drawCFG") {
      DRAW_CFG = true;
    } else if((*i) == "--drawInfPreds") {
      DRAW_INFERRED_PREDS = true;
    } else if((*i) == "--drawPredAbsLTS") {
      DRAW_PRED_ABS_LTS = true;
    } else if((*i) == "--drawFile") {
      if((++i) != commandLine.end()) {
	drawFileName = *i;
	if((drawFileName.size() < 3) || (drawFileName.rfind(".ps") != (drawFileName.size() - 3))) {
	  Util::Error("draw file name \"" + drawFileName + "\" does not end with \".ps\" ...\n");
	}
      } else --i;
    } else if((*i) == "--outputModels") {
      OUTPUT_MODELS = true;
    } else if((*i) == "--modelFile") {
      if((++i) != commandLine.end()) {
	modelFileName = *i;
      } else --i;
    } else if((i->at(0)) == '-') {
      Util::Error("ERROR : unknown option %s ...\n"
		  "use magic --help for usage information ...\n",i->c_str());
    } else{
      FILE_NAMES.push_back(*i);
    }
  }

  //we cannot use lts abstraction with simulation
  assert((CONF_TYPE != CONF_SIMUL) || (!USE_LTS_ABSTRACTION));

  //no need to do incremental verification without LTS abstraction
  INC_VERIFY = USE_LTS_ABSTRACTION ? INC_VERIFY : false;

  //if we are using PAM then we must also use CTOOL
  C_PARSER_TYPE = USE_PAM ? C_PARSER_CTOOL : C_PARSER_TYPE;

  //if we have changed the CEGAR type from its default value then it
  //automatically implies we are doing counter example guided
  //abstraction refinement
  CEGAR = (CEGAR_TYPE != CEGAR_ELIMINATE_PRED) ? true : CEGAR;
  
  //only weak simulation can be checked in explicit state manner. all
  //other types of conformance checks are done using SAT.
  CONF_CHECK_MECHANISM = (CONF_TYPE == CONF_SIMUL) ? CONF_CHECK_MECHANISM : CONF_CHECK_SAT;

  //if we are doing incremental verification we must use SAT for
  //verification and further we must use the HORNSAT solver
  CONF_CHECK_MECHANISM = INC_VERIFY ? CONF_CHECK_SAT : CONF_CHECK_MECHANISM;
  SAT_SOLVER_USED = INC_VERIFY ? SAT_SOLVER_HORN_AI : SAT_SOLVER_USED;

  //for counter-example guided abstraction refinement we need the
  //simulation relation even from SAT
  SIM_REL_FROM_SAT = CEGAR ? true : SIM_REL_FROM_SAT;

  //if we cache impl state transitions then the transitions must be
  //precomputed
  PRECOMPUTE_IMPL_TRANS = CACHE_IMPL_TRANS ? true : PRECOMPUTE_IMPL_TRANS;

  //drawing a more detailed graph implies drawing a less detailed
  //graph
  DRAW_CE_PROJ_LTS = DRAW_CE_PROJ_LTS  || DRAW_CE_PROJ_PRED; 
  DRAW_CE_DAG = DRAW_CE_DAG || DRAW_CE_PROJ_LTS;
  DRAW_COMPOSED_LTS = DRAW_COMPOSED_LTS || DRAW_CE_DAG;
  DRAW_LTS_ABS_LTS = DRAW_LTS_ABS_LTS || DRAW_COMPOSED_LTS;
  DRAW_PRED_ABS_LTS = DRAW_PRED_ABS_LTS || DRAW_LTS_ABS_LTS;
  DRAW_INFERRED_PREDS = DRAW_INFERRED_PREDS || DRAW_PRED_ABS_LTS;
  DRAW_CFG = DRAW_CFG || DRAW_INFERRED_PREDS;
  DRAW_LTS = DRAW_LTS || DRAW_CFG;

  //if we are to draw anything, we must create the output dot file
  CREATE_DOT_FILE = DRAW_LTS;
}

/*********************************************************************/
//parse input files
/*********************************************************************/
void Database::ParseFiles()
{
#ifndef MAGIC_LITE
  Project *proj = NULL;
  if(C_PARSER_TYPE == C_PARSER_CTOOL) {
    proj = new Project();
    proj->SetPrintOption(false);
  }
#endif //MAGIC_LITE
  for(list<string>::iterator i = FILE_NAMES.begin();i != FILE_NAMES.end();i++) {
    //find the filename
    string name = *i;
    //check if the file has been already parsed
    if(IsFileParsed(name)) continue;
    //if it is a preprocessed C file
    if(name.rfind(".pp") == (name.length() - 3)) {
      Util::Message(2,"parsing preprocessed C file %s\n",name.c_str());
      if(C_PARSER_TYPE == C_PARSER_DEFAULT) {
	Stdcin = fopen(name.c_str(),"r");
	//if file could not be opened
	if(Stdcin == NULL) continue;
	//initialise the symbol table and set the file type
	CSymbolTable::fileType = CSymbolTable::STDC_FILE;
	CSymbolTable::Initialise();
	//parse and add the ast to the database if necessary
	Stdcparse();
	assert(CSymbolTable::ast != NULL);
	AddFileAst(name,CSymbolTable::ast);
      } 
#ifndef MAGIC_LITE
      else if(C_PARSER_TYPE == C_PARSER_CTOOL) {
	TransUnit *unit = proj->parse(name.c_str(),false);
	if(USE_PAM) PamAPI::ProcessFile(unit);
	AddFileAst(name,CtoolAPI::ToDefault(unit));
      }
#endif //MAGIC_LITE
      else assert(false);
    }
    //if it is a specification file
    else if(name.rfind(".spec") == (name.length() - 5)) {
      Util::Message(2,"parsing specification file %s\n",name.c_str());      
      Stdcin = fopen(name.c_str(),"r");
      //if file could not be opened
      if(Stdcin == NULL) continue;
      //initialise the symbol table and set the file type
      CSymbolTable::fileType = CSymbolTable::SPEC_FILE;
      CSymbolTable::Initialise();
      //parse and add the name of the spec file to the database
      Stdcparse();
      assert(CSymbolTable::ast == NULL);
      specFiles.insert(name);
    }
  }
#ifndef MAGIC_LITE
  if(USE_PAM) PamAPI::AliasAnalysis();
  if(proj != NULL) delete proj;
#endif //MAGIC_LITE
}

/*********************************************************************/
//returns true iff the file with given name has been parsed already
/*********************************************************************/
bool Database::IsFileParsed(const string &name)
{
  if(name.rfind(".pp") == (name.length() - 3)) {
    return (cAsts.count(name) != 0);
  } else if(name.rfind(".spec") == (name.length() - 5)) {
    return (specFiles.count(name) != 0);
  }
  assert(false);
  return false;
}

/*********************************************************************/
//add a C filename and its ast to the appropriate map
/*********************************************************************/
void Database::AddFileAst(const string &name,StdcFile * ast)
{
  assert(name.rfind(".pp") == (name.length() - 3));
  cAsts[name] = ast;
}

/*********************************************************************/
//extract information about procedures
/*********************************************************************/
void Database::ExtractProcInfos()
{
  //for procedures with code available
  for(map<string,StdcFile*>::iterator i = cAsts.begin();i != cAsts.end();i++) {
    for(list<Proc*>::iterator j = i->second->procs.begin();j != i->second->procs.end();j++) {
      if(procAddInfos.count((*j)->name) == 0) {
	procInfos[(*j)->name] = ProcInfo(**j,i->first);
      } else {
	procInfos[(*j)->name] = ProcInfo(**j,i->first,procAddInfos[(*j)->name]);
      }
    }
  }
  //for procedures without code but only additional info
  for(map<string,ProcAddInfo>::const_iterator i = procAddInfos.begin();i != procAddInfos.end();i++) {
    if(procInfos.count(i->first) == 0) {
      procInfos[i->first] = ProcInfo(i->second);
    }
  }
  //make sure the procedure abstractions are mutually exclusive and
  //complete
  for(map<string,ProcInfo>::iterator i = procInfos.begin();i != procInfos.end();i++) {
    i->second.CheckAbstractions();
  }
}

/*********************************************************************/
//return info about LTS with given name
/*********************************************************************/
const LtsInfo &Database::GetLtsInfo(const string &name)
{
  //first check the global array
  for(size_t i = 0;i < ltsInfos.size();i++) {
    if(ltsInfos[i].GetName() == name) return ltsInfos[i];
  }
  //if not create the lts from the fsp info
  LtsInfo x = fspInfo.GetLtsInfo(name);
  if(DRAW_LTS) x.Draw();
  x.SetIndex(static_cast<char>(ltsInfos.size()));
  ltsInfos.push_back(x);
  return ltsInfos.back();
}

/*********************************************************************/
//return info about negation of the LTS with given name
/*********************************************************************/
const LtsInfo &Database::GetLtsNeg(const string &name)
{
  //if we haven't negated this before
  if(ltsNegMap.count(name) == 0) {
    LtsInfo x = fspInfo.GetLtsNeg(name);
    ltsNegMap[name] = x.GetName();
    x.SetIndex(static_cast<char>(ltsInfos.size()));
    ltsInfos.push_back(x);
    return ltsInfos.back();
  }
  //otherwise just look up the database
  else {
    string negName = ltsNegMap[name];
    for(size_t i = 0;i < ltsInfos.size();i++) {
      if(ltsInfos[i].GetName() == name) return ltsInfos[i];
    }
    assert(false);
  }
}

/*********************************************************************/
//return LTL formula with given name
/*********************************************************************/
LtlFormula Database::GetLtlFormula(const string &name)
{
  return LtlManager::GetFormula(name);
}

/*********************************************************************/
//construct the call graph rooted at the given procedure
/*********************************************************************/
void Database::ConstructCallGraph()
{
  //the list of processes to be handled
  list<string> workSet;
  workSet.insert(workSet.end(),rootProcNames.begin(),rootProcNames.end());
  //the set of processes already handled
  set<string> done;
 
  //process till no more procedures left
  while(!workSet.empty()) {
    string x = workSet.front();
    workSet.pop_front();
    //check if it has been processed already
    if(done.count(x) != 0) continue;
    done.insert(x);
    //get the info about the procedure and simplify it
    if(procInfos.count(x) == 0) {
      Util::Error("procedure " + x + " reachable from but not defined ...\n");
    }
    //get the proc info
    ProcInfo &y = procInfos[x];
    //add a node for this procedure to the call graph
    callGraph.AddNode(x);
    //get the names of procedures called by the current
    //procedure and are not abstracted by actions
    CallGraphCreator cgc(y);
    y.GetAst().Apply(cgc);
    set<string> a = cgc.GetCalledProcs();
    //add these called procedure names to the workset and the edges
    for(set<string>::const_iterator i = a.begin();i != a.end();i++) {
      workSet.push_back(*i);
      callGraph.AddEdge(x,*i);
    }
  }
}

/*********************************************************************/
//update the current time. return the time elapsed since the last
//checkpoint.
/*********************************************************************/
double Database::Checkpoint()
{
#ifdef WIN32
  double newTime = static_cast<double>(time(NULL));
#else
  struct timeval tv;
  if(gettimeofday(&tv,NULL) == -1) {
    Util::Error("could not get current time for checkpointing ...\n");
  }
  double newTime = tv.tv_sec + (tv.tv_usec / 1000000.0);
#endif //WIN32
  Statistics::startTime = (Statistics::startTime == 0) ? newTime : Statistics::startTime;
  Statistics::endTime = newTime;
  double diff = newTime - currTime;
  currTime = newTime;
  return diff;
}

/*********************************************************************/
//return the actions of the specification. the spec could be either a
//FSP process or a LTL formula.
/*********************************************************************/
const set<Action> &Database::GetSpecActions()
{
  if((CONF_TYPE == CONF_SIMUL) || (CONF_TYPE == CONF_REACH)) {
    return GetLtsInfo(specLtsName).GetActions();
  } else if(CONF_TYPE == CONF_LTL) {
    return LtlManager::GetActions(GetLtlFormula(specLtsName));
  } else assert(false);
}

/*********************************************************************/
//return the assignment actions of the specification given a
//context. the spec could be either a FSP process or a LTL formula.
/*********************************************************************/
set<Action> Database::GetSpecAssignActions(const Expr &lhs,const Expr &rhs,
					   const list<string> &params,
					   const set<Expr> &context)
{
  if((CONF_TYPE == CONF_SIMUL) || (CONF_TYPE == CONF_REACH)) {
    return GetLtsInfo(specLtsName).GetAssignActions(lhs,rhs,params,context);
  } else if(CONF_TYPE == CONF_LTL) {
    return GetLtlFormula(specLtsName).GetAssignActions(lhs,rhs,params,context);
  } else assert(false);
}

/*********************************************************************/
//return the assignment actions of the specification given a
//context. the spec could be either a FSP process or a LTL formula.
/*********************************************************************/
set<Action> Database::GetSpecReturnActions(const Expr &expr,const list<string> &params,
					   const set<Expr> &context)
{
  if((CONF_TYPE == CONF_SIMUL) || (CONF_TYPE == CONF_REACH)) {
    return GetLtsInfo(specLtsName).GetReturnActions(expr,params,context);
  } else if(CONF_TYPE == CONF_LTL) {
    return GetLtlFormula(specLtsName).GetReturnActions(expr,params,context);
  } else assert(false);
}

/*********************************************************************/
//return expressions associated with return actions of the
//specification. the spec could be either a FSP process or a LTL
//formula.
/*********************************************************************/
set<Expr> Database::GetSpecRetActExprs()
{
  if((CONF_TYPE == CONF_SIMUL) || (CONF_TYPE == CONF_REACH)) {
    return GetLtsInfo(specLtsName).GetRetActExprs();
  } else if(CONF_TYPE == CONF_LTL) {
    return GetLtlFormula(specLtsName).GetRetActExprs();
  } else assert(false);
}

/*********************************************************************/
//return number of states of the specification automata. the spec
//could be either a FSP process or a LTL formula. for a LTL formula
//return the number of states of the Buchi automaton.
/*********************************************************************/
int Database::GetSpecStateNum()
{
  if((CONF_TYPE == CONF_SIMUL) || (CONF_TYPE == CONF_REACH)) {
    return GetLtsInfo(specLtsName).GetStates().size();
  } else if(CONF_TYPE == CONF_LTL) {
    return Buchi::GetStateNum();
  } else assert(false);
}

/*********************************************************************/
//display usage information for MAGIC
/*********************************************************************/
void Database::DisplayUsage()
{
  Util::Message(1,"\n");
  Util::Message(1,"Usage: magic <option | input file> <option | input file> ...\n");
  Util::Message(1,"\n");
  Util::Message(1,"Common options:\n");
  Util::Message(1,
		"  --help or -h:\n"
		"\tgenerate this help message and exit.\n");
  Util::Message(1,
		"  --parse:\n"
		"\tonly parse input files and exit.\n");
  Util::Message(1,
		"  --echo:\n"
		"\tprint characters as they are being parsed.\n"
		"\tthis is helpful for debugging parse errors.\n");
  Util::Message(1,
		"  --stat:\n"
		"\tdisplay statistics at the end.\n");
  Util::Message(1,
		"  --simFile <file name>:\n"
		"\tname of file to output the computed maximal simulation\n"
		"\trelation. this could be broken.\n");
  Util::Message(1,
		"  --reach:\n"
		"\tcheck ERROR state reachability instead of simulation.\n");
  Util::Message(1,
		"  --trace:\n"
		"\tcheck trace containment instead of simulation.\n");
  Util::Message(1,
		"  --ltl:\n"
		"\tdo SE-LTL model checking. this is still experimental.\n");
  Util::Message(1,
		"  --explicit:\n"
		"\tdo not use HORNSAT for verification. this could be broken.\n");
  Util::Message(1,
		"  --autoPred:\n"
		"\tautomatically add all branches in the code as seed predicates.\n"
		"\tnormally there are no predicates in the first iteration.\n");
  Util::Message(1,
		"  --verbose <level>:\n"
		"\tset the verbosity level. default is 2.\n");
  Util::Message(1,
		"  --noReachImpl:\n"
		"\tdo not perform reachability implicitly. create all possible\n"
		"\tstates. this could increase memory requirement.\n");
  Util::Message(1,
		"  --abstraction <name>:\n"
		"\tthe name of the abstraction to be verified.\n"
		"\tin case of SE-LTL model checking this is the name\n"
		"\tof the formula to be verified.\n");
  Util::Message(1,
		"  --TPCache:\n"
		"\tcache theorem prover queries and their results.\n"
		"\tthis could increase memory requirement.\n");
  Util::Message(1,
		"  --TPCacheSize <size>:\n"
		"\tthe size of the theorem prover cache. the cache is cleared\n"
		"\tif this size is exceeded.\n");
  Util::Message(1,
		"  --optPred:\n"
		"\tdo pseudo-boolean constraint based predicate optimization.\n");
  Util::Message(1,
		"  --cegar:\n"
		"\tdo CEGAR without predicate optimization.\n");
  Util::Message(1,
		"  --ltsAbs:\n"
		"\tdo action-guided abstraction and refinement.\n");
  Util::Message(1,
		"  --noPredAbs:\n"
		"\tdo not do predicate refinement.\n");
  Util::Message(1,
		"  --ceDag <number>:\n"
		"\tuse these many counterexamples in each iteration for refinement.\n");
  Util::Message(1,
		"  --drawLTS:\n"
		"\tdraw all FSP processes used during verification.\n");
  Util::Message(1,
		"  --drawCFG:\n"
		"\tdraw control flow graph for each component. implies --drawLTS.\n");
  Util::Message(1,
		"  --drawInfPreds:\n"
		"\tdraw predicates used at each CFG node in each iteration.\n"
		"\timplies --drawLTS and --drawCFG.\n");
  Util::Message(1,
		"  --drawPredAbsLTS:\n"
		"\tdraw LTS created by predicate abstraction in each iteration.\n"
		"\timplies --drawLTS, --drawCFG and --drawInfPreds.\n");
  Util::Message(1,
		"  --drawFile <file name>:\n"
		"\tname of file where MAGIC will draw. file name must end with .ps.\n"
		"\tdefault is magic.ps.\n");
  Util::Message(1,
		"  --outputModels:\n"
		"\toutput models created in each iteration by predicate abstraction.\n"
		"\tby default, models are output as FSP processes.\n");
  Util::Message(1,
		"  --modelFile <file name>:\n"
		"\tname of file where MAGIC will output models. default is magic.fsp.\n");
  Util::Message(1,"\n");
  Util::Message(1,"Less used options:\n");
  Util::Message(1,
		"  --noImplTrans:\n"
		"\tdo not precompute the transitions of the implementation machine.\n");
  Util::Message(1,
		"  --satSimRel:\n"
		"\tconstruct the simulation relation from the HORNSAT solver's\n"
		"\tinternal data structures. this is automatically set to true\n"
		"\tif some abstraction refinement is to be done.\n");
  Util::Message(1,
		"  --cprover:\n"
		"\tuse CProver as the theorem prover. default is Simplify.\n"); 
  Util::Message(1,
		"  --cvc:\n"
		"\tuse CVC as the theorem prover.\n"); 
  Util::Message(1,
		"  --ics:\n"
		"\tuse ICS as the theorem prover.\n"); 
  Util::Message(1,
		"  --svc:\n"
		"\tuse SVC as the theorem prover.\n"); 
  Util::Message(1,
		"  --sato:\n"
		"\tuse Sato as the SAT solver. default is the internal HORNSAT solver.\n");
  Util::Message(1,
		"  --grasp:\n"
		"\tuse FGrasp as the SAT solver.\n");
  Util::Message(1,
		"  --chaff:\n"
		"\tuse ZChaff as the SAT solver.\n");
  Util::Message(1,
		"  --greedy:\n"
		"\tdo greedy predicate optimization.\n");
  Util::Message(1,
		"  --largestSubset <size>:\n"
		"\tthe maximum size of subsets tried during predicate optimization.\n");
  Util::Message(1,
		"  --smallestSubset <size>:\n"
		"\tthe minimum size of subsets tried during predicate optimization.\n");
  Util::Message(1,
		"  --subsetCutoff <size>:\n"
		"\tthe maximum number of subsets tried during predicate optimization.\n");
  Util::Message(1,
		"  --constraintCutoff <size>:\n"
		"\tthe maximum number of constraints generated during predicate optimization.\n");
  Util::Message(1,
		"  --noStopAfterElim:\n"
		"\tdo not stop generating constraints even if the counterexample\n"
		"\thas been eliminated during predicate optimization.\n");
  Util::Message(1,"\n");
}

/*********************************************************************/
//end of Database.cpp
/*********************************************************************/
