/******************************** CPPFile *****************************

* FileName [GlobalAbsLts.cpp]

* PackageName [main]

* Synopsis [Method definitions of GlobalAbsLts class.]

* SeeAlso [GlobalAbsLts.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <map>
#include <set>
#include <vector>
#include <bitset>
using namespace std;

#include "Node.h"
#include "Action.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "AbsRefiner.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "GlobalAbsLts.h"

/*********************************************************************/
//define static fields
/*********************************************************************/
set<int> GlobalAbsLts::reach;

/*********************************************************************/
//get the set of global initial states
/*********************************************************************/
void GlobalAbsLts::GetInitStates(set<int> &res)
{
  vector< set<int> > absInits;
  for(vector<ProcManager>::const_iterator i = Database::procManagers.begin();
      i != Database::procManagers.end();i++) {
    absInits.push_back(i->GetAbsLts().GetInit());
  }
  CartesianProduct(absInits,0,1,0,res);
}

/*********************************************************************/
//get the set of global initial states in the previous iteration. this
//should only be called if we are doing incremental verification and
//the previous refinement was an LTS refinement.
/*********************************************************************/
void GlobalAbsLts::GetPrevInitStates(set<int> &res)
{
  assert(Database::INC_VERIFY && (Database::lastRefComp != -1));
  vector< set<int> > absInits;
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    if(i == static_cast<size_t>(Database::lastRefComp)) {
      absInits.push_back(Database::procManagers[i].GetAbsLts().GetPrevInit());
    } else {
      absInits.push_back(Database::procManagers[i].GetAbsLts().GetInit());
    }
  }
  CartesianProduct(absInits,0,1,0,res);
}

/*********************************************************************/
//given a vector of sets of component states return the set of
//composite states obtained by their cartesian products
/*********************************************************************/
void GlobalAbsLts::CartesianProduct(const vector< set<int> > &arg,size_t index,int factor,int curr,set<int> &res)
{
  if(index == arg.size()) {
    res.insert(curr);
  } else {
    for(set<int>::const_iterator i = arg[index].begin();i != arg[index].end();i++) {
      CartesianProduct(arg,index + 1,factor * Database::procManagers[index].GetReachImpl().size(),curr + (*i) * factor,res);
    }
  }
}

/*********************************************************************/
//given a global implementation state and an action return the
//possible successors
/*********************************************************************/
void GlobalAbsLts::GetSuccsOnAction(int comp,const Action &act,set<int> &res)
{
  vector<int> absStates = CompToAbsStates(comp);
  vector< set<int> > absSuccs;
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    absSuccs.push_back(Database::procManagers[i].GetAbsLts().GetSuccsOnAction(absStates[i],act));
  }
  CartesianProduct(absSuccs,0,1,0,res);
}

/*********************************************************************/
//given a global implementation state and an action return the
//possible predecessors
/*********************************************************************/
void GlobalAbsLts::GetPredsOnAction(int comp,const Action &act,set<int> &res)
{
  vector<int> absStates = CompToAbsStates(comp);
  vector< set<int> > absPreds;
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    absPreds.push_back(Database::procManagers[i].GetAbsLts().GetPredsOnAction(absStates[i],act));
  }
  CartesianProduct(absPreds,0,1,0,res);
}

/*********************************************************************/
//get the actions
/*********************************************************************/
void GlobalAbsLts::GetActions(set<Action> &res)
{
  for(vector<ProcManager>::const_iterator i = Database::procManagers.begin();
      i != Database::procManagers.end();i++) {
    res.insert(i->GetAbsLts().GetActs().begin(),i->GetAbsLts().GetActs().end());
  }
}

/*********************************************************************/
//get the number of states
/*********************************************************************/
int GlobalAbsLts::GetStateNum()
{
  int res = 1;
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    res *= Database::procManagers[i].GetAbsLts().GetStateNum();
  }
  return res;
}

/*********************************************************************/
//get the number of states in the previous iteration. this should only
//be called if we are doing incremental verification and the previous
//refinement was an LTS refinement.
/*********************************************************************/
int GlobalAbsLts::GetPrevStateNum()
{
  assert(Database::INC_VERIFY && (Database::lastRefComp != -1));
  int res = 1;
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    if(i == static_cast<size_t>(Database::lastRefComp)) {
      res *= Database::procManagers[i].GetAbsLts().GetPrevStateNum();
    } else {
      res *= Database::procManagers[i].GetAbsLts().GetStateNum();
    }
  }
  return res;
}

/*********************************************************************/
//get the set of global states whose successor might have changed due
//to abstraction refinement. this should only be called if we are
//doing incremental verification and the previous refinement was an
//LTS refinement.
/*********************************************************************/
void GlobalAbsLts::GetSuccsChanged(set<int> &res)
{
  assert(Database::INC_VERIFY && (Database::lastRefComp != -1));
  vector< set<int> > absInits;
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    if(i == static_cast<size_t>(Database::lastRefComp)) {
      absInits.push_back(Database::procManagers[i].GetAbsLts().GetSuccsChanged());
    } else {
      absInits.push_back(set<int>());
      for(size_t j = 0;j < Database::procManagers[i].GetAbsLts().GetStateNum();j++) {
	absInits.back().insert(j);
      }
    }
  }
  CartesianProduct(absInits,0,1,0,res);
}

/*********************************************************************/
//given a global implementation state and an action return the
//possible successors in the previous iteration. this should only be
//called if we are doing incremental verification and the previous
//refinement was an LTS refinement.
/*********************************************************************/
void GlobalAbsLts::GetPrevSuccsOnAction(int comp,const Action &act,set<int> &res)
{
  assert(Database::INC_VERIFY && (Database::lastRefComp != -1));
  vector<int> absStates = CompToAbsStates(comp);
  vector< set<int> > absSuccs;
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    if(i == static_cast<size_t>(Database::lastRefComp)) {
      absSuccs.push_back(Database::procManagers[i].GetAbsLts().GetPrevSuccsOnAction(absStates[i],act));
    } else {
      absSuccs.push_back(Database::procManagers[i].GetAbsLts().GetSuccsOnAction(absStates[i],act));
    }
  }
  CartesianProduct(absSuccs,0,1,0,res);
}

/*********************************************************************/
//return the set of propositions true in a global state
/*********************************************************************/
set<Expr> GlobalAbsLts::GetPropositions(int comp)
{
  vector<int> absStates = CompToAbsStates(comp);
  set<Expr> res;
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    const set<Expr> prop = Database::procManagers[i].GetAbsLts().GetPropositions(absStates[i]);
    res.insert(prop.begin(),prop.end());
  }
  return res;
}

/*********************************************************************/
//given a global state id return the vector of component state ids
/*********************************************************************/
vector<int> GlobalAbsLts::CompToAbsStates(int global)
{
  vector<int> res;
  for(vector<ProcManager>::const_iterator i = Database::procManagers.begin();i != Database::procManagers.end();i++) {
    int x = i->GetReachImpl().size();
    res.push_back(global % x);
    global /= x;
  }
  return res;
}

/*********************************************************************/
//end of GlobalAbsLts.cpp
/*********************************************************************/
