/******************************** CPPFile *****************************

* FileName [GlobalCEDag.cpp]

* PackageName [main]

* Synopsis [Method definitions of GlobalCEDag class.]

* SeeAlso [GlobalCEDag.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <set>
#include <map>
#include <vector>
#include <bitset>
using namespace std;

#include "Node.h"
#include "Action.h"
#include "ActionManager.h"
#include "LtsInfo.h"
#include "SimChecker.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "ImplState.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "GlobalAbsLts.h"
#include "LtlChecker.h"
#include "GlobalCEDag.h"

/*********************************************************************/
//define static fields
/*********************************************************************/
vector<int> GlobalCEDag::stateLabel;
vector<Action> GlobalCEDag::actLabel;
vector< set<int> > GlobalCEDag::trans;
set<int> GlobalCEDag::allRoot;
map< int,set< pair<Action,ExclInfo> > > GlobalCEDag::allTrans;
int GlobalCEDag::strategyMax;
int GlobalCEDag::strategyNum;

/*********************************************************************/
//initialise - clear all data structures
/*********************************************************************/
void GlobalCEDag::Initialize()
{
  allRoot.clear();
  allTrans.clear();
}

/*********************************************************************/
//compute the counter example dag from the results of the simulation
//or reachability check
/*********************************************************************/
bool GlobalCEDag::Compute()
{ 
  stateLabel.clear();
  actLabel.clear();
  trans.clear();

  //if checking simulation or ERROR state reachability
  if((Database::CONF_TYPE == Database::CONF_SIMUL) || (Database::CONF_TYPE == Database::CONF_REACH)) {
    return ComputeFsp();
  }
  //if checking LTL formula
  else if(Database::CONF_TYPE == Database::CONF_LTL) {
    return ComputeLtl();
  }
  //impossible
  else assert(false);
}

/*********************************************************************/
//first macro for recursively calling ComputeWithRoot
/*********************************************************************/
inline void GlobalCEDag::MacroOne(const int elemId,const int newNode,bool &newFlag,set<int> &stack,const int loop)
{
  stateLabel.push_back(elemId);
  actLabel.push_back(Action());
  trans.push_back(set<int>());
  stack.insert(newNode);
  ComputeWithRoot(newNode,newFlag,stack,loop);
  stack.erase(newNode);
}

/*********************************************************************/
//second macro. calls the first macro
/*********************************************************************/
inline void GlobalCEDag::MacroTwo(const int node,bool &newFlag,set<int> &stack,const int loop,const ExclInfo &exclInfo,const Action &act)
{
  //update action label and allTrans
  actLabel[node] = act;
  for(set<int>::const_iterator i = exclInfo.first.begin();i != exclInfo.first.end();i++) {
    bool found = false;
#ifdef WIN32
    if((rand() * 1.0 / RAND_MAX) > Database::CE_NEW_NODE_PROB) {
#else
    if((random() * 1.0 / RAND_MAX) > Database::CE_NEW_NODE_PROB) {
#endif //WIN32
      for(size_t l = 0;l < stateLabel.size();l++) {
	if((stack.count(l) == 0) && (stateLabel[l] == (*i))) {
	  trans[node].insert(l);
	  found = true;
	  break;
	}
      }	 
    }   
    if(!found) {
      int newLoop = (exclInfo.second.count(*i) == 0) ? (loop - 1) : loop;
      int newNode = stateLabel.size();
      MacroOne(*i,newNode,newFlag,stack,newLoop);
      trans[node].insert(newNode);
    }
  }
}

/*********************************************************************/
//second macro. calls the first macro
/*********************************************************************/
inline bool GlobalCEDag::MacroThree(const int node,bool &newFlag,set<int> &stack,const int loop,const set<int> &clause,const set<int> &before,const Action &act)
{
  ExclInfo exclInfo(clause,before);
  //check if we can loop
  if((loop == 0) && (!(clause == before))) return false;
  //check if this successor has appeared before
  if(!newFlag) {
    if(IsValidSucc(node,act,clause)) newFlag = true;
    else return false;
  }
  //update action label and allTrans
  allTrans[stateLabel[node]].insert(pair<Action,ExclInfo>(act,exclInfo));
  MacroTwo(node,newFlag,stack,loop,exclInfo,act);
  return true;
}

/*********************************************************************/
//the top-level routine when we are checking for simulation or ERROR
//state reachability
/*********************************************************************/
bool GlobalCEDag::ComputeFsp()
{
  //the spec lts
  const LtsInfo &specInfo = Database::GetLtsInfo(Database::specLtsName);
  //the states of the specification
  const set<string> &specStates = specInfo.GetStates();
  //number of spec lts states
  int ssn = specStates.size();

  //the id of the initial state of the spec lts
  int initSpecId = specInfo.GetStateId(Database::specLtsName);
  //flag to indicate if a new dag was created
  bool newFlag = false;

  //find the first initial state pair that was excluded from the
  //maximal simulation relation and build the CE-DAG with it as the
  //root
  set<int> initImpls; GlobalAbsLts::GetInitStates(initImpls);
  for(set<int>::const_iterator i = initImpls.begin();i != initImpls.end();i++) {
    //if this element was excluded
    int elemId = (*i) * ssn + initSpecId;
    if((allRoot.count(elemId) == 0) && (SimChecker::exclMap.count(elemId) != 0)) {
      newFlag = allRoot.empty();
      allRoot.insert(elemId);
      set<int> stack;
      MacroOne(elemId,stateLabel.size(),newFlag,stack,Database::MAX_CE_LOOP);
      return newFlag;
    }
  }

  //if we reached here it means no new roots were found. this must
  //not be the first attempt.
  assert(!allRoot.empty());
  vector<int> ar = RandomizeIntSet(allRoot);
  newFlag = false;
  set<int> stack;
  MacroOne(ar[0],stateLabel.size(),newFlag,stack,Database::MAX_CE_LOOP);
  return newFlag;
}

/*********************************************************************/
//compute the CE dag with the argument as the root
/*********************************************************************/
void GlobalCEDag::ComputeWithRoot(const int node,bool &newFlag,set<int> &stack,const int loop)
{
  //the spec lts
  const LtsInfo &specInfo = Database::GetLtsInfo(Database::specLtsName);
  //the states of the specification
  const set<string> &specStates = specInfo.GetStates();
  //number of spec lts states
  int ssn = specStates.size();

  //get the abstract state and spec state components
  int absState = stateLabel[node] / ssn;
  int specState = stateLabel[node] % ssn;

  //get the actions of the implementation
  set<Action> iActs; GlobalAbsLts::GetActions(iActs);
  vector<Action> implActs = RandomizeActSet(iActs);

  //check for each possible action and successor
  for(vector<Action>::const_iterator j = implActs.begin();j != implActs.end();j++) {
    set<int> iSuccs; GlobalAbsLts::GetSuccsOnAction(absState,*j,iSuccs);
    vector<int> implSuccs = RandomizeIntSet(iSuccs);
    set<string> specSuccs = specInfo.GetSuccsOnAction(specState,*j);
    for(vector<int>::const_iterator k = implSuccs.begin();k != implSuccs.end();k++) {
      set<int> clause,before;
      bool flag = true;
      for(set<string>::const_iterator m = specSuccs.begin();m != specSuccs.end();m++) {
	int succId = (*k) * ssn + specInfo.GetStateId(*m);
	pair<bool,bool> excl = ExcludedBefore(succId,stateLabel[node]);
	if(excl.first) {
	  clause.insert(succId);
	  if(excl.second) before.insert(succId);
	}
	//for simulation we are done
	else if(Database::CONF_TYPE == Database::CONF_SIMUL) {
	  flag = false;
	  break;
	}
      }

      //if we are checking for simulation
      if(Database::CONF_TYPE == Database::CONF_SIMUL) {
	//check if this is the successor and action to be expanded
	if(flag) {
	  if(MacroThree(node,newFlag,stack,loop,clause,before,*j)) return;
	}
      }
      //else if we are checking for reachability
      else if(Database::CONF_TYPE == Database::CONF_REACH) {
	//try each successor to see if it is suitable
	for(set<int>::const_iterator l = clause.begin();l != clause.end();l++) {
	  set<int> newClause,newBefore;
	  //if the successor has an ERROR component, make it a leaf
	  //node, else extend this path else make
	  if(((*l) % ssn) != specInfo.GetStateId(Database::ERROR_STATE)) {
	    newClause.insert(*l);
	    if(before.count(*l) != 0) newBefore.insert(*l);
	  }
	  if(MacroThree(node,newFlag,stack,loop,newClause,newBefore,*j)) return;
	}
      }
      //this should not happen
      else assert(false);
    }
  }

  //the OR choices
  vector< pair<Action,ExclInfo> > orChoices = RandomizeChoices(allTrans[stateLabel[node]]);
  for(vector< pair<Action,ExclInfo> >::const_iterator j = orChoices.begin();j != orChoices.end();j++) {
    if((loop == 0) && (!(j->second.first == j->second.second))) continue;
    MacroTwo(node,newFlag,stack,loop,j->second,j->first);
    return;
  }
  //this should not happen
  assert(false);
}

/*********************************************************************/
//return true if the first argument was excluded from the maximal
//simulation relation before the second argument and false otherwise
/*********************************************************************/
pair<bool,bool> GlobalCEDag::ExcludedBefore(const int succ,const int pred)
{
  if(SimChecker::exclMap.count(succ) == 0) return pair<bool,bool>(false,false);
  else return pair<bool,bool>(true,SimChecker::exclMap[succ] < SimChecker::exclMap[pred]);
}

/*********************************************************************/
//return true if the choice of successor in the global CE Dag is valid
//- is not the same as the previous choice
/*********************************************************************/
bool GlobalCEDag::IsValidSucc(const int node,const Action &act,const set<int> &succs)
{
  map< int,set< pair<Action,ExclInfo> > >::const_iterator i = allTrans.find(stateLabel[node]);
  if(i == allTrans.end()) return true;

  /*
  //make sure the action is different
  for(set< pair<Action,ExclInfo> >::const_iterator k = i->second.begin();k != i->second.end();k++) {
    if(k->first == act) return false;
  }
  */

  //make sure that the control locations are different
  for(size_t j = 0;j < Database::procManagers.size();j++) {
    set<const ContLoc*> x;
    AbsStatesToContLocs(succs,j,x);    
    bool flag = true;
    for(set< pair<Action,ExclInfo> >::const_iterator k = i->second.begin();k != i->second.end();k++) {
      set<const ContLoc*> y;
      AbsStatesToContLocs(k->second.first,j,y);
      if(x == y) {
	flag = false;
	break;
      }
    }
    if(flag) return true;
  }

  //this choice is inadmissible
  return false;
}

/*********************************************************************/
//given a procedure manager and a set of abstract global
//implementation states return the corresponding set of control
//locations of the procedure manager in the third argument
/*********************************************************************/
void GlobalCEDag::AbsStatesToContLocs(const set<int> &abs,const size_t pmid,set<const ContLoc*> &res)
{
  const LtsInfo &specInfo = Database::GetLtsInfo(Database::specLtsName);
  const set<string> &specStates = specInfo.GetStates();
  int ssn = specStates.size();

  //the set of local abstract states for this proc manager
  set<int> absStates;
  for(set<int>::const_iterator i = abs.begin();i != abs.end();i++) {
    absStates.insert(GlobalAbsLts::CompToAbsStates((*i) / ssn)[pmid]);
  }

  //compute the result
  const ProcManager &pmgr = Database::procManagers[pmid];
  for(set<const ImplState*>::const_iterator i = pmgr.GetReachImpl().begin();i != pmgr.GetReachImpl().end();i++) {
    if(absStates.count((*i)->GetAbsId()) != 0) {
      res.insert((*i)->GetContLoc());
      break;
    }
  }
}

/*********************************************************************/
//the top-level routine when we are checking for LTL property
/*********************************************************************/
bool GlobalCEDag::ComputeLtl()
{
  assert(!LtlChecker::actPath.empty());
  assert(LtlChecker::statePath.size() == (LtlChecker::actPath.size() + 1));
  list<int>::const_iterator i = LtlChecker::statePath.begin();
  list<Action>::const_iterator j = LtlChecker::actPath.begin();
  int loopBack = -1;
  for(;(i !=  LtlChecker::statePath.end()) && (j !=  LtlChecker::actPath.end());i++,j++) {
    size_t newNode = stateLabel.size();
    trans.push_back(set<int>());
    if((*i) ==  LtlChecker::statePath.back()) loopBack = newNode;
    stateLabel.push_back(*i);
    actLabel.push_back(*j);
    j++;
    if(j == LtlChecker::actPath.end()) {
      assert(loopBack != -1);
      trans[newNode].insert(loopBack);
    } else {
      trans[newNode].insert(newNode + 1);
    }
    j--;
  }
  //all done
  return true;
}

/*********************************************************************/
//randomize a set of actions
/*********************************************************************/
vector<Action> GlobalCEDag::RandomizeActSet(const set<Action> &arg)
{
  set<Action> remaining = arg;
  vector<Action> res;
  while(!remaining.empty()) {
#ifdef WIN32
    int randInd = static_cast<int>((rand() * 1.0 / RAND_MAX) * remaining.size());
#else
    int randInd = static_cast<int>((random() * 1.0 / RAND_MAX) * remaining.size());
#endif //WIN32
    int pos = 0;
    for(set<Action>::const_iterator i = remaining.begin();i != remaining.end();i++,pos++) {
      if(pos == randInd) {
	res.push_back(*i);
	remaining.erase(i);
	break;
      }
    }
  }
  assert(res.size() == arg.size());
  return res;
}

/*********************************************************************/
//randomize a set of integers
/*********************************************************************/
vector<int> GlobalCEDag::RandomizeIntSet(const set<int> &arg)
{
  set<int> remaining = arg;
  vector<int> res;
  while(!remaining.empty()) {
#ifdef WIN32
    int randInd = static_cast<int>((rand() * 1.0 / RAND_MAX) * remaining.size());
#else
    int randInd = static_cast<int>((random() * 1.0 / RAND_MAX) * remaining.size());
#endif
    int pos = 0;
    for(set<int>::const_iterator i = remaining.begin();i != remaining.end();i++,pos++) {
      if(pos == randInd) {
	res.push_back(*i);
	remaining.erase(i);
	break;
      }
    }
  }
  assert(res.size() == arg.size());
  return res;
}

/*********************************************************************/
//randomize a set of choices for the impl machine
/*********************************************************************/
vector< pair<Action,ExclInfo> > GlobalCEDag::RandomizeChoices(const set< pair<Action,ExclInfo> > &arg)
{
  set< pair<Action,ExclInfo> > remaining = arg;
  vector< pair<Action,ExclInfo> > res;
  while(!remaining.empty()) {
#ifdef WIN32
    int randInd = static_cast<int>((rand() * 1.0 / RAND_MAX) * remaining.size());
#else
    int randInd = static_cast<int>((random() * 1.0 / RAND_MAX) * remaining.size());
#endif //WIN32
    int pos = 0;
    for(set< pair<Action,ExclInfo> >::const_iterator i = remaining.begin();i != remaining.end();i++,pos++) {
      if(pos == randInd) {
	res.push_back(*i);
	remaining.erase(i);
	break;
      }
    }
  }
  assert(res.size() == arg.size());
  return res;
}

/*********************************************************************/
//end of GlobalCEDag.cpp
/*********************************************************************/
