/******************************** CPPFile *****************************

* FileName [ImplState.cpp]

* PackageName [main]

* Synopsis [Method definitions of ImplState class.]

* SeeAlso [ImplState.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <set>
#include <map>
#include <vector>
#include <typeinfo>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "ProcAbs.h"
#include "Action.h"
#include "ActionManager.h"
#include "LtsTrans.h"
#include "FspInfo.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ProcInfo.h"
#include "LtsInfo.h"
#include "ImplState.h"
#include "ContLoc.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"

/*********************************************************************/
//methods of ImStSet
/*********************************************************************/
ImStSet::ImStSet(const set<const ImplState*> &e,const ProcManager &procManager) 
{
  ele = e;
  all = (ele.size() == procManager.GetReachImpl().size());
  if(all) ele.clear();
}

const ImStSet &ImStSet::operator = (const ImStSet &rhs)
{
  all = rhs.all;
  ele = rhs.ele;
  return *this;
}

bool ImStSet::operator == (const ImStSet &rhs) const
{
  return ((all == rhs.all) && (ele == rhs.ele));
}

bool ImStSet::operator < (const ImStSet &rhs) const
{
  if(all && !rhs.all) return false;
  else if(!all && rhs.all) return true;
  else return (ele < rhs.ele);
}

/*********************************************************************/
//return the set of elements given the proc manager
/*********************************************************************/
const set<const ImplState*> &ImStSet::GetEle(const ProcManager &procManager) const
{
  return all ? procManager.GetReachImpl() : ele;
}

/*********************************************************************/
//get pre-image under an action
/*********************************************************************/
ImStSet ImStSet::GetPredsOnAction(const Action &act,const ProcManager &procManager) const
{
  const set<const ImplState*> &allEle = GetEle(procManager);
  set<const ImplState*> res;
  for(set<const ImplState*>::const_iterator i = allEle.begin();i != allEle.end();i++) {
    set<const ImplState*> x = (*i)->GetPredsOnAction(act);
    res.insert(x.begin(),x.end());
  }
  return ImStSet(res,procManager);      
}

/*********************************************************************/
//get post-image under an action
/*********************************************************************/
ImStSet ImStSet::GetSuccsOnAction(const Action &act,const ProcManager &procManager) const
{
  const set<const ImplState*> &allEle = GetEle(procManager);
  set<const ImplState*> res;
  for(set<const ImplState*>::const_iterator i = allEle.begin();i != allEle.end();i++) {
    set<const ImplState*> x = (*i)->GetSuccsOnAction(act);
    res.insert(x.begin(),x.end());
  }
  return ImStSet(res,procManager);      
}

/*********************************************************************/
//return a set of all impl states
/*********************************************************************/
ImStSet ImStSet::GetComplete()
{
  ImStSet res;
  res.all = true;
  return res;
}

/*********************************************************************/
//intersect this set with the argument set
/*********************************************************************/
void ImStSet::Intersect(const ImStSet &rhs)
{
  if(all) *this = rhs;
  else {
    set<const ImplState*> newEle;
    for(set<const ImplState*>::iterator i = ele.begin();i != ele.end();i++) {
      if(rhs.all || (rhs.ele.count(*i) != 0)) newEle.insert(*i);
    }
    ele = newEle;
  }
}

/*********************************************************************/
//pick an element
/*********************************************************************/
const ImplState *ImStSet::Pick(const ProcManager &procManager) const
{
  if(all) return *(procManager.GetReachImpl().begin());
  else {
    assert(!ele.empty());
    return *(ele.begin());
  }
}

/*********************************************************************/
//return true if this set contains the argument
/*********************************************************************/
bool ImStSet::Contains(const ImplState *arg) const
{
  return (all || (ele.count(arg) != 0));
}

bool ImStSet::Contains(const ContLoc *arg) const
{
  if(all) return true;
  for(set<const ImplState*>::const_iterator i = ele.begin();i != ele.end();i++) {
    if(arg == (*i)->GetContLoc()) return true;
  }
  return false;
}

/*********************************************************************/
//returns true if this set is empty
/*********************************************************************/
bool ImStSet::Empty() const
{
  return ((!all) && (ele.empty()));
}

/*********************************************************************/
//display a set of impl states
/*********************************************************************/
void ImStSet::Display(const ProcManager &procManager) const
{
  Util::Message(2,"************** begin impl state set **************\n");
  const set<const ImplState*> &is = all ? procManager.GetReachImpl() : ele;
  for(set<const ImplState*>::const_iterator i = is.begin();i != is.end();i++) {
    Util::Message(2,"%s\n",(*i)->ToString().c_str());
  } 
  Util::Message(2,"*************** end impl state set ***************\n");
}

/*********************************************************************/
//constructors
/*********************************************************************/
ImplState::ImplState(short c,const PVal &p,short ss,char si)
{
  contLocId = c;
  predVal = p;
  specStateId = ss;
  specInfoId = si;
  absId = -1;
}

ImplState::ImplState(const ImplState &rhs)
{
  *this = rhs;
}

/*********************************************************************/
//operators
/*********************************************************************/
const ImplState &ImplState::operator = (const ImplState &rhs)
{
  contLocId = rhs.contLocId;
  predVal = rhs.predVal;
  specStateId = rhs.specStateId;
  specInfoId = rhs.specInfoId;
  absId = rhs.absId;
  return *this;
}

/*********************************************************************/
//return the set of specification propositions true in this state
/*********************************************************************/
set<Expr> ImplState::GetPropositions() const
{
  return GetContLoc()->GetPropositions(GetKey());
}

/*********************************************************************/
//return the set of successors of this state alongwith the
//corresponding actions
/*********************************************************************/
list< pair<const ImplState*,Action> > ImplState::GetSuccsAndActions() const
{
  //first check if the successors and actions have been already
  //computed before
  ContLoc *contLoc = GetContLoc();
  string key = GetKey();
  if(Database::PRECOMPUTE_IMPL_TRANS) {
    if(contLoc->ContainsSuccsAndActions(key)) {
      return contLoc->GetSuccsAndActions(key);
    }
  }
  
  list< pair<const ImplState*,Action> > ret;
  //last control location
  if(contLoc->GetLocType() == ContLoc::FINAL) {
    ret = GetSuccsAndActionsFinal();
  }
  //return location
  else if(contLoc->GetLocType() == ContLoc::RETURN) {
    ret = GetSuccsAndActionsReturn();
  }
  //branch statement
  else if(contLoc->GetLocType() == ContLoc::BRANCH) {
    ret = GetSuccsAndActionsBranch();
  }
  //assignment statement with rhs not a procedure call
  else if(contLoc->GetLocType() == ContLoc::ASSIGN_EXPR) {
    ret = GetSuccsAndActionsAssign();
  }
  //assignment statement with rhs being a procedure call
  else if(contLoc->GetLocType() == ContLoc::ASSIGN_CALL) {
    ret = GetSuccsAndActionsCall();
  }
  //unknown 
  else assert(false);
  
  //update the control location map
  if(Database::PRECOMPUTE_IMPL_TRANS) {
    contLoc->AddSuccsAndActions(key,ret);
  }
  
  //return the result
  return ret;
}

/*********************************************************************/
//return the set of successors of this state on the argument action
/*********************************************************************/
set<const ImplState*> ImplState::GetSuccsOnAction(const Action &act) const
{
  set<const ImplState*> res;
  list< pair<const ImplState*,Action> > allSuccs = GetSuccsAndActions();
  for(list< pair<const ImplState*,Action> >::const_iterator i = allSuccs.begin();i != allSuccs.end();i++) {
    if(i->second == act) res.insert(i->first);
  }
  return res;
}

/*********************************************************************/
//return the set of predecessors of this state on the argument action
/*********************************************************************/
set<const ImplState*> ImplState::GetPredsOnAction(const Action &act) const
{
  set<const ImplState*> res;
  list< pair<const ImplState*,Action> > allPreds = GetContLoc()->GetPredsAndActions(GetKey());
  for(list< pair<const ImplState*,Action> >::const_iterator i = allPreds.begin();i != allPreds.end();i++) {
    if(i->second == act) res.insert(i->first);
  }
  return res;
}

/*********************************************************************/
//return the set of actions exported by this state
/*********************************************************************/
set<Action> ImplState::GetActions() const
{
  ContLoc *contLoc = GetContLoc();
  set<Expr> state = contLoc->GetPreds().ToExpr(contLoc->GetPreds().PValToPredVal(predVal));

  //the result
  set<Action> res;

  //last control location
  if(contLoc->GetLocType() == ContLoc::FINAL) {
    if(Database::CONF_TYPE == Database::CONF_LTL) {
      res.insert(ActionManager::GetImplEpsilonAction(GetProcManagerId()));
    }
  }
  //return location
  else if(contLoc->GetLocType() == ContLoc::RETURN) {
    const Expr &expr = contLoc->GetLocExpr();
    res = Database::GetSpecReturnActions(expr,contLoc->GetProcInfo()->GetParams(),state);
    if(res.empty()) res.insert(ActionManager::GetDefaultRetAction());
  }
  //branch statement
  else if(contLoc->GetLocType() == ContLoc::BRANCH) {
    res.insert(ActionManager::GetImplEpsilonAction(GetProcManagerId()));
  }
  //assignment statement with rhs not a procedure call
  else if(contLoc->GetLocType() == ContLoc::ASSIGN_EXPR) {
    const Expr &lhs = contLoc->GetLocLhs();
    const Expr &rhs = contLoc->GetLocRhs();
    res = Database::GetSpecAssignActions(lhs,rhs,contLoc->GetProcInfo()->GetParams(),state);
    if(res.empty()) res.insert(ActionManager::GetImplEpsilonAction(GetProcManagerId()));
  }
  //assignment statement with rhs being a procedure call
  else if(contLoc->GetLocType() == ContLoc::ASSIGN_CALL) {
    Expr lhs = contLoc->GetLocLhs();
    Expr rhs = contLoc->GetLocRhs();
    
    //find the list of arguments to the call
    pair< Expr,list<Expr> > comps = ExprManager::GetCalledProcDetails(rhs);
    list<Expr> &args = comps.second;
    
    //we have to inline the abstraction LTSs. find the successors and
    //actions of the spec state associated.
    LtsInfo &specInfo = Database::ltsInfos[specInfoId];
    string ss = specInfo.GetIdState(specStateId);
    list<LtsTrans> outTrans = specInfo.GetOutTrans(ss);
    
    //add the implementation state for the destination state
    //of each outgoing transition
    for(list<LtsTrans>::iterator i = outTrans.begin();i != outTrans.end();i++) {
      string fs = i->GetFinalState();
      //if the destination state is a STOP state we have to
      //consider the effect of return actions
      if(fs == Database::STOP_STATE) {
	Action x = i->GetAction();
	assert(x.IsReturn());
	Expr retExpr = x.GetConcreteReturnExpr(args);
	//if the return expression is void
	if(retExpr.IsEmptyExpr()) {
	  res.insert(ActionManager::GetImplEpsilonAction(GetProcManagerId()));
	}
	//if this is a default return
	else if(retExpr.ToString() == "! ") {
	  res.insert(ActionManager::GetDefaultRetAction());
	}
	//if the return expression is non-void and non-default
	else {
	  set<Action> acts = Database::GetSpecAssignActions(lhs,retExpr,contLoc->GetProcInfo()->GetParams(),state);
	  if(acts.empty()) acts.insert(Action(lhs,retExpr));
	  res.insert(acts.begin(),acts.end());
	}
      }
      //otherwise the final state is not stop
      else {
	Action x = i->GetAction();
	assert(!x.IsReturn());
	//otherwise if this is an assign action
	if(x.IsAssign()) {
	  pair<Expr,Expr> concAssign = x.GetConcreteAssignExprs(args);
	  set<Action> acts = Database::GetSpecAssignActions(concAssign.first,concAssign.second,
							    contLoc->GetProcInfo()->GetParams(),state);
	  if(acts.empty()) acts.insert(Action(concAssign.first,concAssign.second));	  
	  res.insert(acts.begin(),acts.end());
	}
	//otherwise this is a basic action
	else {
	  //if this is a spec epsilon action we change it to a impl epsilon action
	  if(x.IsSpecEpsilon()) x = ActionManager::GetImplEpsilonAction(GetProcManagerId());
	  res.insert(x);
	}
      }
    }
  }
  //unknown 
  else assert(false);

  return res;
}

/*********************************************************************/
//create the key for this implementation state by appending the
//predicate valuation, a colon and the inlined spec LTS state. the key
//is used to map the successors and actions of the implementation
//state in the map associated with the control location for this
//implementation state.
/*********************************************************************/
string ImplState::GetKey() const
{
  char x[100],y[100];
  snprintf(x,100,"%lu",PredSet::PValToUL(predVal));
  if(specStateId == -1) {
    return string(x) + ":null";
  } else {
    snprintf(y,100,":%d:%d",specStateId,(short)(specInfoId));
    return string(x) + string(y);
  }
}

/*********************************************************************/
//return successors of the final control location
/*********************************************************************/
list< pair<const ImplState*,Action> > ImplState::GetSuccsAndActionsFinal() const
{
  //if we are doing ltl model checking and this is the first parallel
  //component then allow the final location to stutter
  if(Database::CONF_TYPE == Database::CONF_LTL) {
    list< pair<const ImplState*,Action> > res;
    if(GetProcManagerId() == 0) {
      set<Action> acts; acts.insert(ActionManager::GetImplEpsilonAction(0));
      AddSuccAndActions(res,this,acts);
    }
    return res;
  }
  //otherwise the final location has no successors. if there are any
  //return successors they will be handled by the return location
  else if((Database::CONF_TYPE == Database::CONF_SIMUL) || (Database::CONF_TYPE == Database::CONF_REACH)) {   
    return list< pair<const ImplState*,Action> >();
  } else assert(false);
}

/*********************************************************************/
//create the successors and actions when the next control locations is
//in the same procedure and there is an additional constraint apart
//from those due to the predicate valuation. the first argument is the
//successor control location. the second argument is the constraint. a
//NULL value indicates that there is no additional constraint. the
//third argument is the set of actions. the fourth argument is the
//list where the results are collected.
/*********************************************************************/
void ImplState::HandleConstraint(ContLoc *succ,const Expr &cons,const set<Action> &acts,
				 list< pair<const ImplState*,Action> > &ret) const
{
  const ContLoc *contLoc = GetContLoc();
  const PredSet &preds = contLoc->GetPreds();
  set<Expr> state = preds.ToExpr(preds.PValToPredVal(predVal));
  
  set<Expr> negList;
  if(!cons.IsNull()) negList.insert(ExprManager::NegateExpr(cons));
  
  if(!ExprManager::ProveImplies(state,negList)) {
    if(!cons.IsNull()) state.insert(cons);
    const PredSet &succPreds = succ->GetPreds();
    
    //get the set of possible predicate valuations in the successor
    //assuming there is no assignment.
    vector<PVal> pvals = succPreds.GetPossibleValuations(preds,predVal,list<Expr>(),list<Expr>());
    
    for(size_t i = 0;i < pvals.size();i++) {
      const PVal &pval = pvals[i];
      //get the expression for the successor state
      set<Expr> expr = succPreds.ToExpr(succPreds.PValToPredVal(pval));
      //negate the expressions
      set<Expr>  neg = ExprManager::NegateExprSet(expr);
      //if the constraints cannot imply this negation then we
      //have to consider this predicate valuation as a possible
      //successor
      if(!ExprManager::ProveImplies(state,neg)) {
	set<const ImplState*> a = succ->GetImplStates(pval);
	for(set<const ImplState*>::const_iterator j = a.begin();j != a.end();j++) {
	  AddSuccAndActions(ret,*j,acts);
	}
      }
    }
  }
}

/*********************************************************************/
//create the successors and actions when the next control locations is
//in the same procedure and there is a parallel assignment. the first
//argument is the successor control location. the second and third
//arguments are the lhs-list and rhs-list of the assignment. the
//fourth argument is the set of actions. the fifth argument is the
//list where the results are collected. the last argument (if not
//equal to -1), indicates the id of the spec state to be associated
//with the successor states.
/*********************************************************************/
void ImplState::HandleAssign(ContLoc *succ,const list<Expr> &lhsList,const list<Expr> &rhsList,
			     const set<Action> &acts,list< pair<const ImplState*,Action> > &ret,
			     const short fsid,const short fiid) const
{
  ContLoc *contLoc = GetContLoc();  
  //first compute the set of possible valuation of the successors
  PredSet &preds = contLoc->GetPreds();
  PredSet &succPreds = succ->GetPreds();
  vector<PVal> pvals = succPreds.GetPossibleValuations(preds,predVal,lhsList,rhsList);
  
  //for each possible predicate valuation of the successor check
  //if it is a possible next state valuation. if so add the
  //corresponding successor impl state.
  for(size_t i = 0;i < pvals.size();i++) {
    const PVal &pval = pvals[i];
    if(PossibleNextValuation(lhsList,rhsList,preds,predVal,succPreds,pval)) {
      if(fsid == -1) {
	set<const ImplState*> implStates = succ->GetImplStates(pval);
	for(set<const ImplState*>::const_iterator j = implStates.begin();j != implStates.end();j++) {
	  AddSuccAndActions(ret,*j,acts);
	}
      } else {
	AddSuccAndActions(ret,succ->GetImplState(pval,fsid,fiid),acts);
      }
    }
  }
}

/*********************************************************************/
//return successors of a return control location. 
/*********************************************************************/
list< pair<const ImplState*,Action> > ImplState::GetSuccsAndActionsReturn() const
{
  list< pair<const ImplState*,Action> > ret;

  //get the successor location. this should be a final location.
  ContLoc *contLoc = GetContLoc();
  ContLoc *fin = contLoc->GetTSucc();
  assert(fin->GetLocType() == ContLoc::FINAL);
  
  //get the return expression
  const Expr &expr = contLoc->GetLocExpr();

  //get the concrete expression for the return state
  set<Expr> state = contLoc->GetPreds().ToExpr(contLoc->GetPreds().PValToPredVal(predVal));
  
  //the list of actions associated
  set<Action> acts;

  //the program will terminate with this return
  acts = Database::GetSpecReturnActions(expr,contLoc->GetProcInfo()->GetParams(),state);
  if(acts.empty()) acts.insert(ActionManager::GetDefaultRetAction());
  HandleConstraint(fin,Expr(),acts,ret);
  
  //return the final result
  return ret;
}

/*********************************************************************/
//return the successors of a branching statement
/*********************************************************************/
list< pair<const ImplState*,Action> > ImplState::GetSuccsAndActionsBranch() const
{
  ContLoc *contLoc = GetContLoc();
  const Expr &branchCond = contLoc->GetLocExpr();
  Expr falseCond = ExprManager::NegateExpr(branchCond);
  set<Action> acts; acts.insert(ActionManager::GetImplEpsilonAction(GetProcManagerId()));
  list< pair<const ImplState*,Action> > ret;

  HandleConstraint(contLoc->GetTSucc(),branchCond,acts,ret);
  HandleConstraint(contLoc->GetESucc(),falseCond,acts,ret);

  return ret;
}

/*********************************************************************/
//return the successors of an assignment statement where the rhs is
//not a procedure call
/*********************************************************************/
list< pair<const ImplState*,Action> > ImplState::GetSuccsAndActionsAssign() const
{
  ContLoc *contLoc = GetContLoc();

  //get the lhs and rhs of the assignment
  const Expr &lhs = contLoc->GetLocLhs();
  const Expr &rhs = contLoc->GetLocRhs();
  
  //create the aliases set of the lhs
  set<Expr> lhsAliases = contLoc->GetLhsAliases();
  lhsAliases.insert(lhs);

  //find the set of actions associated with the assignment
  set<Expr> state = contLoc->GetPreds().ToExpr(contLoc->GetPreds().PValToPredVal(predVal));
  set<Action> acts;
  for(set<Expr>::const_iterator i = lhsAliases.begin();i != lhsAliases.end();i++) {
    set<Action> x = Database::GetSpecAssignActions(*i,rhs,contLoc->GetProcInfo()->GetParams(),state);
    acts.insert(x.begin(),x.end());
  }
  if(acts.empty()) acts.insert(ActionManager::GetImplEpsilonAction(GetProcManagerId()));

  //the final result
  list< pair<const ImplState*,Action> > ret;

  //do all the stuff
  list<Expr> lhsList,rhsList; 
  for(set<Expr>::const_iterator i = lhsAliases.begin();i != lhsAliases.end();i++) {
    lhsList.push_back(*i);
    rhsList.push_back(rhs);
  }
  HandleAssign(contLoc->GetTSucc(),lhsList,rhsList,acts,ret);

  //all done
  return ret;	
}

/*********************************************************************/
//return the successors of an call statement. first argument is the
//lvalue where the return value is stored. second argument is the
//actual call.
/*********************************************************************/
list< pair<const ImplState*,Action> > ImplState::GetSuccsAndActionsCall() const
{
  ContLoc *contLoc = GetContLoc();
  Expr lhs = contLoc->GetLocLhs();
  Expr rhs = contLoc->GetLocRhs();
  
  //get the successor control location
  ContLoc *succ = contLoc->GetTSucc();

  //the return list
  list< pair<const ImplState*,Action> > ret;
  
  //get the concrete expression for the current state
  set<Expr> state = contLoc->GetPreds().ToExpr(contLoc->GetPreds().PValToPredVal(predVal));

  //find the list of arguments to the call
  pair< Expr,list<Expr> > comps = ExprManager::GetCalledProcDetails(rhs);
  list<Expr> &args = comps.second;

  //we have to inline the abstraction LTSs. find the successors and
  //actions of the spec state associated.
  LtsInfo &specInfo = Database::ltsInfos[specInfoId];
  string ss = specInfo.GetIdState(specStateId);
  list<LtsTrans> outTrans = specInfo.GetOutTrans(ss);
  
  //add the implementation state for the destination state
  //of each outgoing transition
  for(list<LtsTrans>::const_iterator i = outTrans.begin();i != outTrans.end();i++) {
    string fs = i->GetFinalState();
    //if the destination state is a STOP state we have to
    //consider the effect of return actions
    if(fs == Database::STOP_STATE) {
      Action x = i->GetAction();
      assert(x.IsReturn());
      Expr retExpr = x.GetConcreteReturnExpr(args);
      //if the return expression is void
      if(retExpr.IsEmptyExpr()) {
	set<Action> acts; acts.insert(ActionManager::GetImplEpsilonAction(GetProcManagerId()));
	HandleConstraint(succ,Expr(),acts,ret);
      }
      //if this is a default return
      else if(retExpr.ToString() == "! ") {
	set<Action> acts; acts.insert(ActionManager::GetDefaultRetAction());
	HandleConstraint(succ,Expr(),acts,ret);
      }
      //if the return expression is non-void and non-default
      else {
	set<Action> acts = Database::GetSpecAssignActions(lhs,retExpr,contLoc->GetProcInfo()->GetParams(),state);
	if(acts.empty()) acts.insert(Action(lhs,retExpr));
	list<Expr> lhsList; lhsList.push_back(lhs);
	list<Expr> rhsList; rhsList.push_back(retExpr);
	HandleAssign(succ,lhsList,rhsList,acts,ret);
      }
    }
    //otherwise the final state is not stop
    else {
      Action x = i->GetAction();
      assert(!x.IsReturn());
      short fsid = specInfo.GetStateId(fs);
      //otherwise if this is an assign action
      if(x.IsAssign()) {
	pair<Expr,Expr> concAssign = x.GetConcreteAssignExprs(args);
	set<Action> acts = Database::GetSpecAssignActions(concAssign.first,concAssign.second,
							  contLoc->GetProcInfo()->GetParams(),state);
	if(acts.empty()) acts.insert(Action(concAssign.first,concAssign.second));	  
	list<Expr> lhsList; lhsList.push_back(concAssign.first);
	list<Expr> rhsList; rhsList.push_back(concAssign.second);
	HandleAssign(contLoc,lhsList,rhsList,acts,ret,fsid,specInfoId);	  
      }
      //otherwise this is a basic action
      else {
	//if this is a spec epsilon action we change it to a impl epsilon action
	if(x.IsSpecEpsilon()) x = ActionManager::GetImplEpsilonAction(GetProcManagerId());
	set<Action> acts; acts.insert(x);
	const ImplState *sis = contLoc->GetImplState(predVal,fsid,specInfoId);
	AddSuccAndActions(ret,sis,acts);
      }
    }
  }

  return ret;
}

/*********************************************************************/
//add a list of successors and actions to a list. the successor is the
//second argument. the list of actions is the third argument. for each
//action in the list create an object array of size 2. make the first
//element the successor and the second element the action. add this
//array to the first argument.
/*********************************************************************/
void ImplState::AddSuccAndActions(list< pair<const ImplState*,Action> > &ret,const ImplState *succ,const set<Action> &actions) const
{
  for(set<Action>::const_iterator i = actions.begin();i != actions.end();i++) {
    pair<const ImplState*,Action> x;
    x.first = succ;
    x.second = *i;
    ret.push_back(x);
    Util::Message(3,ToString() + " === " + i->ToString() + " ==> " + succ->ToString() + "\n");
  }
}

/*********************************************************************/
//string representation
/*********************************************************************/
string ImplState::ToString() const
{
  ContLoc *contLoc = GetContLoc();
  //get the expression for the predicate valuation and convert it to a string
  set<Expr> a = contLoc->GetPreds().ToExpr(contLoc->GetPreds().PValToPredVal(predVal));
  string b = ExprManager::ExprSetToString(a);
  if(specStateId == -1) {
    return (contLoc->ToString() + ":" + "null" + ":" + b);
  } else {
    return (contLoc->ToString() + ":" + Database::ltsInfos[specInfoId].GetIdState(specStateId) + ":" + b);
  }
}

/*********************************************************************/
//check if a new predicate valuation is possible from an old valuation
//due to an assignment statement.
/*********************************************************************/
bool ImplState::PossibleNextValuation(const list<Expr> &lhsList,const list<Expr> &rhsList,
				      const PredSet &oldPreds,const PVal &oldVal,
				      const PredSet &newPreds,const PVal &newVal) const
{
  //get the concrete expression for the old valuation
  set<Expr> oldExpr = oldPreds.ToExpr(oldPreds.PValToPredVal(oldVal));
  
  //get the concrete expression for the new valuation
  set<Expr> newExpr1 = newPreds.ToExpr(newPreds.PValToPredVal(newVal));
  
  //get the negation of the weakest preconditions of the
  //concrete new valuation expressions
  set<Expr> newExpr2;
  for(set<Expr>::const_iterator i = newExpr1.begin();i != newExpr1.end();i++) {
    Expr a = ExprManager::ComputeWP(lhsList,rhsList,*i);
    newExpr2.insert(ExprManager::NegateExpr(a));
  }

  //finally check implication
  return !ExprManager::ProveImplies(oldExpr,newExpr2);
}

/*********************************************************************/
//return the index of the proc manager of this impl state
/*********************************************************************/
size_t ImplState::GetProcManagerId() const
{
  return (contLocId % Database::rootProcNames.size());
}

/*********************************************************************/
//return the pointer to the control location for this state
/*********************************************************************/
ContLoc *ImplState::GetContLoc() const
{
  //the proc manager index
  size_t pmi = contLocId % Database::rootProcNames.size();
  //the cont loc index
  size_t cli = contLocId / Database::rootProcNames.size();
  return Database::procManagers[pmi].GetUsefulLocs()[cli];
}

/*********************************************************************/
//end of ImplState.cpp
/*********************************************************************/
