/******************************** CPPFile *****************************

* FileName [InputProcessor.cpp]

* PackageName [main]

* Synopsis [This file contains method definitions of the
* InputProcessor class.]

* SeeAlso [InputProcessor.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <map>
#include <vector>
#include <set>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "LtsInfo.h"
#include "DFSAdapter.h"
#include "ProgAbs.h"
#include "ProgInfo.h"
#include "Action.h"
#include "ProcAbs.h"
#include "ProcInfo.h"
#include "LtsTrans.h"
#include "FspInfo.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "CallGraph.h"
#include "ContLocExtractor.h"
#include "ProcInliner.h"
#include "AbsRefiner.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "TheoremProver.h"
#include "LtlFormula.h"
#include "Buchi.h"
#include "LtlToBuchi.h"
#include "InputProcessor.h"

/*********************************************************************/
//do all the stuff
/*********************************************************************/
bool InputProcessor::Run()
{
  //parse command line arguments and check if there is some
  //abstraction to be verified
  Database::ParseCommandLine();

  //initialise the theorem prover
  TheoremProver::Initialise();

  //open the output dot file if necessary
  if(Database::CREATE_DOT_FILE) {
    assert(Database::dotFilePtr == NULL);
    Database::dotFilePtr = fopen(Database::DOT_FILE_NAME.c_str(),"w");
    assert(Database::dotFilePtr != NULL);    
    fprintf(Database::dotFilePtr,"digraph magic_dot_graph {\n\n");
    fprintf(Database::dotFilePtr,"  node [color=blue];\n");
    fprintf(Database::dotFilePtr,"  edge [color=red];\n\n");
  }

  //open the output model file if necessary
  if(Database::OUTPUT_MODELS) {
    assert(Database::modelFilePtr == NULL);
    Database::modelFilePtr = fopen(Database::modelFileName.c_str(),"w");
    assert(Database::modelFilePtr != NULL);    
  }

  //parse the input files and return if this is all to be done
  Database::ParseFiles();
  if(Database::PARSE_ONLY) return false;

  //if we need to simply translate ltl formulae to buchi automata
  if(Database::LTL_TO_BUCHI) {
    LtlToBuchi::Run();
    return false;
  }

  //if no abstractions to be verified
  if(Database::ABSTRACTION_NAME == "") return false;

  //add epsilon self-loops if not checking LTL formula
  if(Database::CONF_TYPE != Database::CONF_LTL) Database::fspInfo.AddEpsilonTrans();

  //extract the information about the procedures
  Database::ExtractProcInfos();

  //get the abstraction that needs to be checked
  for(map<string,ProgInfo>::const_iterator i = Database::progInfos.begin();
      i != Database::progInfos.end();i++) {
    const list<ProgAbs> &pabs = i->second.GetAbstractions();
    for(list<ProgAbs>::const_iterator j = pabs.begin();j != pabs.end();j++) {
      if(Database::ABSTRACTION_NAME == j->GetName()) {       
	Database::specLtsName = j->GetLtsName();

	//convert trace containment to reachability
	if(Database::CONF_TYPE == Database::CONF_TRACE) {
	  Database::CONF_TYPE = Database::CONF_REACH;
	  const LtsInfo &negLts = Database::GetLtsNeg(Database::specLtsName);
	  Database::specLtsName = negLts.GetName();
	}

	//initialise the buchi automaton
	if(Database::CONF_TYPE == Database::CONF_LTL) {
	  int res = Buchi::Initialize(Database::GetLtlFormula(Database::specLtsName));
	  if(res == Buchi::LTL_TAUTOLOGY) {
	    Util::Error("LTL formula is a tautology ...\n");
	  } else if(res == Buchi::LTL_UNSATISFIABLE) {
	    Util::Error("LTL formula is unsatisfiable ...\n");
	  }
	  Buchi::Display();
	  Util::Message(2,"Buchi initialised ...\n");
	}

	Database::progName = i->first;
	Database::rootProcNames = i->second.GetProcs();
	const vector<Expr> absGuards = j->GetGuards();
	assert(Database::rootProcNames.size() == absGuards.size());
	for(size_t k = 0;k < Database::rootProcNames.size();k++) {
	  ProcInfo &pi = Database::procInfos[Database::rootProcNames[k]];
	  Database::absGuards.push_back(CreateMainPred(pi,absGuards[k]));
	}
	Util::Message(2,"main predicates created ...\n");	
	
	//compute relevant ltl propositions
	if(Database::CONF_TYPE == Database::CONF_LTL) {
	  for(size_t k = 0;k < Database::rootProcNames.size();k++) {
	    Database::procInfos[Database::rootProcNames[k]].ComputeProps();
	  }
	  Util::Message(2,"relevant propositions computed ...\n");
	}

	//construct the call graph for the root procedures
	Database::ConstructCallGraph();
	
	//get the nodes of the call graph
	const set<string> &scc = Database::callGraph.GetNodes();
	
	//if there are no procedures in the call graph we are done
	if(scc.empty()) return false;

	//create the control locations for the procedures
	for(set<string>::const_iterator k = scc.begin();k != scc.end();k++) {
	  //create the control locations
	  ProcInfo &y = Database::procInfos[*k];
	  assert(y.GetContLocs().empty());
	  ContLocExtractor cle(*k);
	  y.GetAst().Apply(cle);
	  //y.DisplayContLocs();
	}
	Util::Message(2,"control locations created ...\n");

	//remove unnecessary control locations - cases, defaults,
	//labels, gotos and breaks. these act as skip locations.
	RemoveSkipLocations();
	Util::Message(2,"skip locations eliminated ...\n");
	
	//remove useless branches i.e. branches whose true and false
	//successors are the same
	RemoveUselessBranches();
	Util::Message(2,"useless branch locations eliminated ...\n");

	//inline procedure bodies at unabstracted call sites and
	//associate info about spec LTSs to abstracted call sites
	ProcInliner::Inline();
	Util::Message(2,"procedures inlined ...\n");
	
	//create the managers for the root procedures. also create the
	//array of all control locations for each root procedure and
	//backup fields and compute predecessors.
	for(size_t k = 0;k < Database::rootProcNames.size();k++) {
	  Database::procManagers.push_back(ProcManager(k,Database::rootProcNames[k],Database::absGuards[k]));
	  Database::procManagers[k].Initialise();
	  Database::procManagers[k].CreateAllContLocs();
	  Util::Message(2,"procedure manager for " + Database::rootProcNames[k] + " created ...\n");
	}

	//free control locations in the procedure infos
	for(map<string,ProcInfo>::iterator k = Database::procInfos.begin();
	    k != Database::procInfos.end();k++) {
	  k->second.Cleanup();
	}

	//do the live variable analysis for the control locations
	DoLiveVarAnalysis();
	Util::Message(2,"live variable analysis done ...\n");

	//if we are doing cegar build the initial maps
	if(Database::CEGAR && Database::USE_PRED_ABSTRACTION) {
	  BuildInitialMaps();
	  Util::Message(2,"initial maps built ...\n");
	}	
	//else seed the control locations with the necessary
	//predicates
	else {
	  ComputeNecessaryPredicates();
	  Util::Message(2,"necessary predicates computed ...\n");
	}
	
	return true;
      }
    }
  }

  //abstraction to be checked not found
  return false;
}

/*********************************************************************/
//create the concrete main predicate by replacing dummy variables with
//parameters
/*********************************************************************/
Expr InputProcessor::CreateMainPred(ProcInfo &mainInfo,const Expr &mainPred)
{
  //replace the occurences of "$1", "$2" etc with the parameters
  //of the main procedure
  list<Expr> repList = ExprManager::StringListToExprList(mainInfo.GetParams());
  return ExprManager::ReplaceDummyVars(mainPred,repList,1);
}

/*********************************************************************/
//remove unnecessary control locations - cases, defaults, labels,
//gotos and breaks. these act as skip locations and need not be
//analysed.
/*********************************************************************/
void InputProcessor::RemoveSkipLocations()
{
  const set<string> &scc = Database::callGraph.GetNodes();
  
  //redirect the predecessors of the skip locations to the
  //successors of the skip locations.
  for(set<string>::const_iterator i = scc.begin();i != scc.end();i++) {
    ProcInfo &b = Database::procInfos[*i];

    //set the initial location
    if(b.GetInitLoc()->IsSkipLoc()) {
      b.SetInitLoc(GetClosestNonSkipLoc(b.GetInitLoc()));
    }

    for(list<ContLoc*>::iterator j = b.GetContLocs().begin();j != b.GetContLocs().end();j++) {
      ContLoc *c = *j;
      if(c->IsSkipLoc()) continue;
      if((c->GetLocType() == ContLoc::FINAL) || (c->GetLocType() == ContLoc::RETURN)) {
      } else if(c->GetLocType() == ContLoc::ASSIGN_EXPR) {
	c->SetTSucc(GetClosestNonSkipLoc(c->GetTSucc()));
      } else if(c->GetLocType() == ContLoc::ASSIGN_CALL) {
	c->SetTSucc(GetClosestNonSkipLoc(c->GetTSucc()));
	ContLoc *d = c->GetCallSucc();
	if(d != NULL) {
	  c->SetCallSucc(GetClosestNonSkipLoc(d));
	}
      } else if(c->GetLocType() == ContLoc::BRANCH) {
	c->SetTSucc(GetClosestNonSkipLoc(c->GetTSucc()));
	c->SetESucc(GetClosestNonSkipLoc(c->GetESucc()));
      } else Util::Error("control location of unknown type ...\n");
    }
  }
  //now remove all skip locations
  for(set<string>::const_iterator i = scc.begin();i != scc.end();i++) {
    ProcInfo &b = Database::procInfos[*i];
    for(list<ContLoc*>::iterator j = b.GetContLocs().begin();j != b.GetContLocs().end();) {
      if((*j)->IsSkipLoc()) {
	delete (*j);
	j = b.GetContLocs().erase(j);
      } else j++;
    }
  }
}

/*********************************************************************/
//given a control location, if it is non-skip return it. otherwise
//return the non-skip location closest to it. if there is a loop of
//skip locations including this one then it throws an error.
/*********************************************************************/
ContLoc *InputProcessor::GetClosestNonSkipLoc(ContLoc *arg)
{
  set<ContLoc*> seen;
  while(true) {
    //if this is non-skip
    if(!arg->IsSkipLoc()) return arg;
    //otherwise check for skip loop
    if(seen.count(arg) != 0) {
      Util::Error("skip location loop detected ...\n");
    }
    seen.insert(arg);
    arg = arg->GetTSucc();
  }
}

/*********************************************************************/
//eliminate branch locations whose true and false successors are
//identical
/*********************************************************************/
void InputProcessor::RemoveUselessBranches()
{
  const set<string> &scc = Database::callGraph.GetNodes();
 
  for(set<string>::const_iterator i = scc.begin();i != scc.end();i++) {
    ProcInfo &b = Database::procInfos[*i];
    while(true) {
      //get a useless branch
      ContLoc* ulb = NULL;
      for(list<ContLoc*>::iterator j = b.GetContLocs().begin();j != b.GetContLocs().end();j++) {
	if(((*j)->GetLocType() == ContLoc::BRANCH) && ((*j)->GetTSucc() == (*j)->GetESucc())) {
	  assert((*j) != (*j)->GetTSucc());
	  ulb = *j;
	  break;
	}
      }

      //if there are no useless branches we are done
      if(ulb == NULL) break;

      //set the initial location
      if(b.GetInitLoc() == ulb) {
	b.SetInitLoc(b.GetInitLoc()->GetTSucc());
      }

      //redirect all predecessors of useless branches to the
      //successors
      for(list<ContLoc*>::iterator j = b.GetContLocs().begin();j != b.GetContLocs().end();j++) {
	ContLoc *loc = (*j);
	if(loc == ulb) continue;
	int locType = loc->GetLocType();
	if((locType == ContLoc::FINAL) || (locType == ContLoc::RETURN)) {
	} else if(locType == ContLoc::ASSIGN_EXPR) {
	  if(loc->GetTSucc() == ulb) {
	    loc->SetTSucc(loc->GetTSucc()->GetTSucc());
	    assert(loc != loc->GetTSucc());
	  }
	} else if(locType == ContLoc::ASSIGN_CALL) {
	  if(loc->GetTSucc() == ulb) {
	    loc->SetTSucc(loc->GetTSucc()->GetTSucc());
	    assert(loc != loc->GetTSucc());
	  }
	  if(loc->GetCallSucc() == ulb) {
	    loc->SetCallSucc(loc->GetCallSucc()->GetTSucc());
	    assert(loc != loc->GetCallSucc());
	  }
	} else if(locType == ContLoc::BRANCH) {
	  if(loc->GetTSucc() == ulb) {
	    loc->SetTSucc(loc->GetTSucc()->GetTSucc());
	    assert(loc != loc->GetTSucc());
	  }
	  if(loc->GetESucc() == ulb) {
	    loc->SetESucc(loc->GetESucc()->GetTSucc());
	    assert(loc != loc->GetESucc());
	  }
	} else assert(false);
      }

      //remove all the useless branches
      for(list<ContLoc*>::iterator j = b.GetContLocs().begin();j != b.GetContLocs().end();j++) {
	if((*j) == ulb) {
	  delete (*j);
	  b.GetContLocs().erase(j);
	  break;
	}
      }
    }
  }
}

/*********************************************************************/
//do the live variable analysis for each root procedure
/*********************************************************************/
void InputProcessor::DoLiveVarAnalysis()
{
  for(vector<ProcManager>::const_iterator i = Database::procManagers.begin();
      i != Database::procManagers.end();i++) {
    i->DoLiveVariableAnalysis();
  }
}

/*********************************************************************/
//initializes the trace mappings etc.
/*********************************************************************/
void InputProcessor::BuildInitialMaps()
{
  for(vector<ProcManager>::iterator i = Database::procManagers.begin();
      i != Database::procManagers.end();i++) {
    i->BuildInitialMaps();
  }
}

/*********************************************************************/
//seed the control locations with necessary predicates
/*********************************************************************/
void InputProcessor::ComputeNecessaryPredicates()
{
  for(vector<ProcManager>::const_iterator i = Database::procManagers.begin();
      i != Database::procManagers.end();i++) {
    i->ComputeNecessaryPredicates();
  }
}

/*********************************************************************/
//end of InputProcessor.cpp
/*********************************************************************/
