/******************************** CPPFile *****************************

* FileName [LocalAbsLts.cpp]

* PackageName [main]

* Synopsis [Method definitions of LocalAbsLts class.]

* SeeAlso [LocalAbsLts.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cassert>
#include <string>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "Action.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "ImplState.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"

/*********************************************************************/
//operators
/*********************************************************************/
const LocalAbsLts &LocalAbsLts::operator = (const LocalAbsLts &rhs)
{
  pmid = rhs.pmid;
  stateNum = rhs.stateNum;
  prevStateNum = rhs.prevStateNum;
  init = rhs.init;
  prevInit = rhs.prevInit;
  propMap = rhs.propMap;
  acts = rhs.acts;
  succsChanged = rhs.succsChanged;
  succMap = rhs.succMap;
  predMap = rhs.predMap;
  prevSuccMap = rhs.prevSuccMap;
  return *this;
}

/*********************************************************************/
//return the set of propositions true in the argument state
/*********************************************************************/
const set<Expr> LocalAbsLts::GetPropositions(int state) const
{
  map< int,set<Expr> >::const_iterator i = propMap.find(state);
  return (i == propMap.end()) ? set<Expr>() : i->second;
}

/*********************************************************************/
//given an abstract state and an action return the successor abstract
//states
/*********************************************************************/
set<int> LocalAbsLts::GetSuccsOnAction(int from,const Action &act)
{
  //if the action is not exported by this machine then stutter
  if(acts.count(act) == 0) {
    set<int> x; x.insert(from);
    return x;
  } 
  //otherwise look at the successor transition relation
  else {
    //if successors of this state have not been computed
    if(succMap.count(from) == 0) {
      const ProcManager &pmgr = Database::procManagers[pmid];
      for(set<const ImplState*>::const_iterator i = pmgr.GetReachImpl().begin();i != pmgr.GetReachImpl().end();i++) {
       	if((*i)->GetAbsId() == from) {
       	  list< pair<const ImplState*,Action> > sa = (*i)->GetSuccsAndActions();
       	  for(list< pair<const ImplState*,Action> >::const_iterator k = sa.begin();k != sa.end();k++) {
       	    succMap[from][k->second].insert(k->first->GetAbsId());
       	    predMap[k->first->GetAbsId()][k->second].insert(from);
       	  }
       	}
      }
    } 
    map< int,map< Action,set<int> > >::const_iterator i = succMap.find(from);
    if(i == succMap.end()) return set<int>();
    map< Action,set<int> >::const_iterator j = i->second.find(act);
    if(j == i->second.end()) return set<int>();
    else return j->second;
  }
}

/*********************************************************************/
//given an abstract state and an action return the predecessor
//abstract states
/*********************************************************************/
set<int> LocalAbsLts::GetPredsOnAction(int from,const Action &act) const
{
  //if the action is not exported by this machine then stutter
  if(acts.count(act) == 0) {
    set<int> x; x.insert(from);
    return x;
  } 
  //otherwise look at the predecessor transition relation
  else {
    map< int,map< Action,set<int> > >::const_iterator i = predMap.find(from);
    if(i == predMap.end()) return set<int>();
    map< Action,set<int> >::const_iterator j = i->second.find(act);
    if(j == i->second.end()) return set<int>();
    else return j->second;
  }
}

/*********************************************************************/
//given an abstract state and an action return the successor abstract
//states in the previous iteration
/*********************************************************************/
set<int> LocalAbsLts::GetPrevSuccsOnAction(int from,const Action &act)
{
  //if the action is not exported by this machine then stutter
  if(acts.count(act) == 0) {
    set<int> x; x.insert(from);
    return x;
  } 
  //otherwise look at the successor transition relation
  else {
    map< int,map< Action,set<int> > >::const_iterator i = prevSuccMap.find(from);
    if(i == prevSuccMap.end()) return set<int>();
    map< Action,set<int> >::const_iterator j = i->second.find(act);
    if(j == i->second.end()) return set<int>();
    return j->second;
  }
}

/*********************************************************************/
//display
/*********************************************************************/
void LocalAbsLts::Display()
{
  Util::Message(2,"******************* state ******************\n");
  const ProcManager &pmgr = Database::procManagers[pmid];
  for(size_t j = 0;j < stateNum;j++) {
    for(set<const ImplState*>::const_iterator k = pmgr.GetReachImpl().begin();k != pmgr.GetReachImpl().end();k++) {
      if((*k)->GetAbsId() == (int)(j)) Util::Message(2,"%d : %s\n",j,(*k)->ToString().c_str());
    }
  }
  Util::Message(2,"******************* trans ******************\n");
  for(set<Action>::const_iterator i = acts.begin();i != acts.end();i++) {
    for(size_t j = 0;j < stateNum;j++) {
      set<int> x = GetSuccsOnAction(j,*i);
      for(set<int>::const_iterator k = x.begin();k != x.end();k++) {
       	Util::Message(2,"%d === %s ===> %d\n",j,i->ToString().c_str(),*k);
      }
    }
  }
  Util::Message(2,"********************************************\n");
}

/*********************************************************************/
//end of LocalAbsLts.cpp
/*********************************************************************/
