/******************************** CPPFile *****************************

* FileName [LtsAbsRefiner.cpp]

* PackageName [main]

* Synopsis [Method definitions of LtsAbsRefiner class.]

* SeeAlso [LtsAbsRefiner.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <vector>
#include <set>
#include <map>
#include <list>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "Action.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "ImplState.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "LtsAbsRefiner.h"

/*********************************************************************/
//define static fields
/*********************************************************************/
const int LtsAbsRefiner::LTS_ABS_REFINED = 11700;
const int LtsAbsRefiner::LTS_ABS_NOT_REFINED = 11710;

/*********************************************************************/
//do it all
/*********************************************************************/
int LtsAbsRefiner::Run()
{
  //create the old map from implStates to absIds
  map<const ImplState*,int> oldAbsIds;
  for(set<const ImplState*>::const_iterator i = procManager.reachImpl.begin();i != procManager.reachImpl.end();i++) {
    oldAbsIds[*i] = (*i)->GetAbsId();
  }

  //refine
  LocalAbsLts &absLts = procManager.absLts;
  absLts.prevStateNum = absLts.stateNum;
  //ids of split states
  set<int> currIds; 
  for(map< int,set<Action> >::const_iterator it = procManager.absLtsRefine.begin();it != procManager.absLtsRefine.end();it++) {
    //split specified abstract state on specified action
    map< vector< set<int> >,set<ImplState*> > succMap;      
    for(set<const ImplState*>::const_iterator i = procManager.reachImpl.begin();i != procManager.reachImpl.end();i++) {
      if(oldAbsIds[*i] == it->first) {
       	vector< set<int> > succVec;
       	for(set<Action>::const_iterator m = it->second.begin();m != it->second.end();m++) {
       	  set<const ImplState*> soa = (*i)->GetSuccsOnAction(*m);
       	  set<int> aoa;
       	  for(set<const ImplState*>::const_iterator k = soa.begin();k != soa.end();k++) {
       	    aoa.insert(oldAbsIds[*k]);
       	  }
       	  succVec.push_back(aoa);
       	}
       	succMap[succVec].insert((ImplState*)(*i));
      }
    }

    if(succMap.size() > 1) {
      //update the ids of the implementation states and proposition map
      for(map< vector< set<int> >,set<ImplState*> >::const_iterator i = succMap.begin();i != succMap.end();i++) {
       	//update the ids
       	if(i == succMap.begin()) continue;
       	for(set<ImplState*>::const_iterator j = i->second.begin();j != i->second.end();j++) {
       	  (*j)->SetAbsId(absLts.stateNum);
       	  currIds.insert(oldAbsIds[*j]);
	  currIds.insert((*j)->GetAbsId());
       	}
       	//update the proposition map
       	absLts.propMap[absLts.stateNum] = (*(i->second.begin()))->GetPropositions();
       	absLts.stateNum++;
      } 
    } 
  }
  
  //compute new abstract initial states
  Util::Message(2,"number of abstract implementation states = %d\n",absLts.stateNum);
  absLts.prevInit = absLts.init;
  absLts.init.clear();
  set<const ImplState*> initImpls = procManager.initLoc->GetImplStates(procManager.guard);
  for(set<const ImplState*>::const_iterator i = initImpls.begin();i != initImpls.end();i++) {
    absLts.init.insert((*i)->GetAbsId());
  }

  //compute the new abstract successors and actions
  if(Database::PRECOMPUTE_IMPL_TRANS) {
    absLts.succsChanged = currIds;
    //ids of predecessors of split states
    set<int> predIds;
    for(set<int>::const_iterator i = currIds.begin();i != currIds.end();i++) {
      const map< Action,set<int> > &x = absLts.predMap[*i];
      for(map< Action,set<int> >::const_iterator j = x.begin();j != x.end();j++) {
	for(set<int>::const_iterator k = j->second.begin();k != j->second.end();k++) {
	  if(currIds.count(*k) == 0) {
	    predIds.insert(*k);
	    absLts.succsChanged.insert(*k);
	  }
	}	
      }
    }
    //ids of successors of split states
    set<int> succIds;
    for(set<int>::const_iterator i = currIds.begin();i != currIds.end();i++) {
      const map< Action,set<int> > &x = absLts.succMap[*i];
      for(map< Action,set<int> >::const_iterator j = x.begin();j != x.end();j++) {
	for(set<int>::const_iterator k = j->second.begin();k != j->second.end();k++) {
	  if(currIds.count(*k) == 0) {
	    succIds.insert(*k);
	  }
	}
      }
    }
    //ids of predecessors of successors of split states
    set<int> succPredIds;
    for(set<int>::const_iterator i = succIds.begin();i != succIds.end();i++) {
      const map< Action,set<int> > &x = absLts.predMap[*i];
      for(map< Action,set<int> >::const_iterator j = x.begin();j != x.end();j++) {
	for(set<int>::const_iterator k = j->second.begin();k != j->second.end();k++) {
	  if(currIds.count(*k) == 0) {
	    succPredIds.insert(*k);
	  }
	}
      }
    }

    //save the necessary transition relations from previous iteration
    absLts.prevSuccMap.clear();
    for(set<int>::const_iterator i = absLts.succsChanged.begin();i != absLts.succsChanged.end();i++) {
      absLts.prevSuccMap[*i] = absLts.succMap[*i];
    }

    //clear predecessors and successors that will be recomputed
    for(set<int>::const_iterator i = currIds.begin();i != currIds.end();i++) {
      absLts.succMap[*i].clear();
      absLts.predMap[*i].clear();
    }
    for(set<int>::const_iterator i = predIds.begin();i != predIds.end();i++) {
      absLts.succMap[*i].clear();
    }
    for(set<int>::const_iterator i = succIds.begin();i != succIds.end();i++) {
      absLts.predMap[*i].clear();
    }
    for(set<int>::const_iterator i = succPredIds.begin();i != succPredIds.end();i++) {
      absLts.succMap[*i].clear();
    }

    //map from abstract id to impl states
    map< int,set<const ImplState*> > idToImpls;
    for(set<const ImplState*>::const_iterator i = procManager.reachImpl.begin();i != procManager.reachImpl.end();i++) {
      idToImpls[(*i)->GetAbsId()].insert(*i);
    }

    //recompute successors of predecessor states
    for(set<int>::const_iterator i = predIds.begin();i != predIds.end();i++) {
      for(set<const ImplState*>::const_iterator j = idToImpls[*i].begin();j != idToImpls[*i].end();j++) {
      	list< pair<const ImplState*,Action> > sa = (*j)->GetSuccsAndActions();
	for(list< pair<const ImplState*,Action> >::const_iterator k = sa.begin();k != sa.end();k++) {
	  absLts.succMap[(*j)->GetAbsId()][k->second].insert(k->first->GetAbsId());
	  absLts.predMap[k->first->GetAbsId()][k->second].insert((*j)->GetAbsId());
	}
      }
    }
    //recompute successors of current states
    for(set<int>::const_iterator i = currIds.begin();i != currIds.end();i++) {
      for(set<const ImplState*>::const_iterator j = idToImpls[*i].begin();j != idToImpls[*i].end();j++) {
       	list< pair<const ImplState*,Action> > sa = (*j)->GetSuccsAndActions();
       	for(list< pair<const ImplState*,Action> >::const_iterator k = sa.begin();k != sa.end();k++) {
       	  absLts.succMap[(*j)->GetAbsId()][k->second].insert(k->first->GetAbsId());
       	  absLts.predMap[k->first->GetAbsId()][k->second].insert((*j)->GetAbsId());
       	}
      }
    }
    //recompute successors of predecessors of successor states
    for(set<int>::const_iterator i = succPredIds.begin();i != succPredIds.end();i++) {
      for(set<const ImplState*>::const_iterator j = idToImpls[*i].begin();j != idToImpls[*i].end();j++) {
       	list< pair<const ImplState*,Action> > sa = (*j)->GetSuccsAndActions();
       	for(list< pair<const ImplState*,Action> >::const_iterator k = sa.begin();k != sa.end();k++) {
       	  absLts.succMap[(*j)->GetAbsId()][k->second].insert(k->first->GetAbsId());
       	  absLts.predMap[k->first->GetAbsId()][k->second].insert((*j)->GetAbsId());
       	}
      }
    }
  }
  procManager.absLtsRefine.clear();
  //all done
  return (absLts.stateNum > absLts.prevStateNum) ? LTS_ABS_REFINED : LTS_ABS_NOT_REFINED;
}

/*********************************************************************/
//end of LtsAbsRefiner.cpp
/*********************************************************************/
