/******************************** CPPFile *****************************

* FileName [ModelExtractor.cpp]

* PackageName [main]

* Synopsis [This file contains method definitions of the
* ModelExtractor class.]

* SeeAlso [ModelExtractor.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <set>
#include <map>
#include <vector>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Node.h"
#include "Action.h"
#include "ProcAbs.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "ProcInfo.h"
#include "CallGraph.h"
#include "AbsRefiner.h"
#include "LocalAbsLts.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "ModelExtractor.h"

/*********************************************************************/
//extract finite state model from C program. return true on success
//and false otherwise.
/*********************************************************************/
bool ModelExtractor::Run()
{
  //clear all caches of the expression manager
  ExprManager::ClearCaches();

  //compute the implementation states
  ComputeImplStates();
  Util::Message(2,"implementation states for control locations computed ...\n");

  //compute the successors and actions if required
  if(Database::PRECOMPUTE_IMPL_TRANS) {
    ComputeSuccsAndActions();
    Util::Message(2,"successors and actions of control locations computed ...\n");
  }
  return true;
}

/*********************************************************************/
//clear predicate sets at all control locations
/*********************************************************************/
void ModelExtractor::ClearPredSets()
{
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    Database::procManagers[i].ClearPredSets();
  }
}

/*********************************************************************/
//compute the set of implementation states for each control location
/*********************************************************************/
void ModelExtractor::ComputeImplStates()
{
  //compute the implementation states
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    if(Database::procManagers[i].GetNewSeeds()) {
      Database::procManagers[i].ComputeImplStates();      
      if(Database::DRAW_INFERRED_PREDS) Database::procManagers[i].DrawInferredPreds();
      if(Database::DRAW_PRED_ABS_LTS) Database::procManagers[i].DrawPredAbsLts();
      if(Database::OUTPUT_MODELS) Database::procManagers[i].OutputModel();
      Database::procManagers[i].CreateInitialAbstractStates();
      Database::procManagers[i].ComputeAbsLtsInit();
      Database::procManagers[i].SetNewSeeds(false);
    }
  }
}

/*********************************************************************/
//compute the successors and actions of control locations. also create
//the info about the abstract LTSs.
/*********************************************************************/
void ModelExtractor::ComputeSuccsAndActions()
{
  for(size_t i = 0;i < Database::procManagers.size();i++) {
    Database::procManagers[i].ComputeSuccsAndActions();
    Database::procManagers[i].ComputeAbsLtsTrans();
  }
}

/*********************************************************************/
//end of ModelExtractor.cpp
/*********************************************************************/
