/******************************** CPPFile *****************************

* FileName [ProcManager.cpp]

* PackageName [main]

* Synopsis [Definitions for ProcManager class.]

* SeeAlso [ProcManager.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string>
#include <vector>
#include <list>
#include <set>
#include <map>
#include <bitset>
using namespace std;

#include "Util.h"
#include "Statistics.h"
#include "Node.h"
#include "LtsInfo.h"
#include "Action.h"
#include "ActionManager.h"
#include "ProcInfo.h"
#include "LtlFormula.h"
#include "LtlManager.h"
#include "Database.h"
#include "Predicate.h"
#include "PredSet.h"
#include "ContLoc.h"
#include "ImplState.h"
#include "AbsRefiner.h"
#include "LocalAbsLts.h"
#include "GlobalCEDag.h"
#include "ConcCEDag.h"
#include "ProcManager.h"
#include "StdcParser.h"

/*********************************************************************/
//constructors and destructors
/*********************************************************************/
ProcManager::ProcManager() { assert(false); }

ProcManager::ProcManager(size_t i,const string &n,const Expr &g) : name(n),guard(g),absLts(i)
{ 
  index = i;
  newSeeds = true; 
  clauseNum = 0; 

  //create the file names
  char prefix[20];
  snprintf(prefix,20,"pbs-%d-%d",getpid(),index);
  constraintFile = string(prefix) + ".con";
  dimacsFile = string(prefix) + ".cnf";
  pbFile = string(prefix) + ".cnf.pb";
  outputFile = string(prefix) + ".out";
}

ProcManager::ProcManager(const ProcManager &rhs) { *this = rhs; }

/*********************************************************************/
//operators
/*********************************************************************/
const ProcManager &ProcManager::operator = (const ProcManager &rhs)
{
  index = rhs.index;
  name = rhs.name;
  guard = rhs.guard;
  allLocs = rhs.allLocs;
  branchLocs = rhs.branchLocs;
  usefulLocs = rhs.usefulLocs;
  initLoc = rhs.initLoc;
  initSkipped = rhs.initSkipped;
  finalLoc = rhs.finalLoc;
  reachImpl = rhs.reachImpl;
  optDags = rhs.optDags;
  locToPreds = rhs.locToPreds;
  locToUseful = rhs.locToUseful;
  necPreds = rhs.necPreds;
  necUseful = rhs.necUseful;
  assignToBranch = rhs.assignToBranch;
  seedBranches = rhs.seedBranches;
  newSeeds = rhs.newSeeds;
  absLts = rhs.absLts;
  varMap = rhs.varMap;
  mapVar = rhs.mapVar;
  allCons = rhs.allCons;
  dagToLocs = rhs.dagToLocs;
  clauseNum = rhs.clauseNum;
  constraintFile = rhs.constraintFile;
  dimacsFile = rhs.dimacsFile;
  pbFile = rhs.pbFile;
  outputFile = rhs.outputFile;
  seedMap = rhs.seedMap;
  absLtsRefine = rhs.absLtsRefine;
  return *this;
}

/*********************************************************************/
//initialise the abstraction refiner
/*********************************************************************/
void ProcManager::Initialise() const
{
  //remove any old copies of the data files
  remove(constraintFile.c_str());
  remove(dimacsFile.c_str());
  remove(pbFile.c_str());
  remove(outputFile.c_str());
}

/*********************************************************************/
//cleanup the abstraction refiner
/*********************************************************************/
void ProcManager::Cleanup() const
{
  //remove any current copies of the data files
  remove(constraintFile.c_str());
  remove(dimacsFile.c_str());
  remove(pbFile.c_str());
  remove(outputFile.c_str());

  //remove any impl state and transition files
  for(map< set<ContLoc*>,int >::const_iterator i = seedMap.begin();i != seedMap.end();i++) {
    char buf[32];
    snprintf(buf,32,"implstate-%d-%d-%d",getpid(),index,i->second);
    remove(buf);
    snprintf(buf,32,"impltrans-%d-%d-%d",getpid(),index,i->second);
    remove(buf);
  }
}

/*********************************************************************/
//create the vector of all control locations. also backup fields and
//compute predecessors.
/*********************************************************************/
void ProcManager::CreateAllContLocs()
{
  //make a copy
  map<ContLoc*,ContLoc*> oldToNew;
  const list<ContLoc*> &oldLocs = Database::procInfos[name].GetContLocs();
  for(list<ContLoc*>::const_iterator i = oldLocs.begin();i != oldLocs.end();i++) {
    ContLoc *x = new ContLoc(**i);
    allLocs.push_back(x);
    oldToNew[*i] = x;
  }

  //change the successors
  for(vector<ContLoc*>::const_iterator i = allLocs.begin();i != allLocs.end();i++) {
    int locType = (*i)->GetLocType();
    if(locType == ContLoc::FINAL) {
    } else if((locType == ContLoc::RETURN) || (locType == ContLoc::ASSIGN_EXPR)) {
      (*i)->SetTSucc(oldToNew[(*i)->GetTSucc()]);
    } else if(locType == ContLoc::ASSIGN_CALL) {
      assert((*i)->GetCallSucc() == NULL);
      (*i)->SetTSucc(oldToNew[(*i)->GetTSucc()]);
    } else if(locType == ContLoc::BRANCH) {
      (*i)->SetTSucc(oldToNew[(*i)->GetTSucc()]);
      (*i)->SetESucc(oldToNew[(*i)->GetESucc()]);
    } else assert(false);
  }

  //set the initial and final locations
  initLoc = oldToNew[Database::procInfos[name].GetInitLoc()];
  finalLoc = oldToNew[Database::procInfos[name].GetFinalLoc()];

  //backup and compute predecessors. also create the set of all branch
  //locations.
  for(vector<ContLoc*>::iterator i = allLocs.begin();i != allLocs.end();i++) {
    if((*i)->GetLocType() == ContLoc::BRANCH) branchLocs.insert(*i);
    (*i)->Backup();
    (*i)->MakePred();
  }

  //draw the control flow graph if necessary
  if(Database::DRAW_CFG) DrawCFG();
}

/*********************************************************************/
//perform live variable analysis
/*********************************************************************/
void ProcManager::DoLiveVariableAnalysis() const
{
  //initialise the analysis and create the working set
  set<ContLoc*> wset;
  for(size_t i = 0;i < allLocs.size();i++) {
    allLocs[i]->InitLiveVarAnalysis();
    wset.insert(allLocs[i]);
  }
  
  //then do the analysis
  while(!wset.empty()) {
    ContLoc *a = *(wset.begin());
    wset.erase(wset.begin());
    if(a->UpdateLiveVars()) {
      const set<ContLoc*> &x = a->GetPredecessors();
      wset.insert(x.begin(),x.end());
    }
  }
}

/*********************************************************************/
//initializes the trace mappings etc.
/*********************************************************************/
void ProcManager::BuildInitialMaps()
{
  //create the set of locations to optimize on. these are the set of
  //branch locations in the program.
  set<ContLoc*> optLocs;
  for(vector<ContLoc*>::const_iterator i = allLocs.begin();i != allLocs.end();i++) {
    if ((*i)->GetLocType() == ContLoc::BRANCH)
      optLocs.insert(*i);
  }
  Util::Message(2,"Number of locations to be optimized : %d\n",optLocs.size());
  
  //collect the set of predicates at each control for each branch
  //location
  for(set<ContLoc*>::const_iterator i = optLocs.begin();i != optLocs.end ();i++) {
    ClearPredSets();
    (*i)->AddPred(Predicate((*i)->GetLocExpr(),Predicate::REQUIRED,pair<const ContLoc*,int>((*i),1)));
    SetupLocsNoCegar();
    for(vector<ContLoc*>::const_iterator j = allLocs.begin();
      j != allLocs.end ();j++) {
      locToPreds[*i][*j] = (*j)->GetPreds();
    }
    locToUseful[*i].insert(usefulLocs.begin(),usefulLocs.end());
  }

  //create the map for necessary predicates
  ClearPredSets();
  ComputeNecessaryPredicates();
  SetupLocsNoCegar();
  for(vector<ContLoc*>::const_iterator i = allLocs.begin();
      i != allLocs.end ();i++) {
    necPreds[*i] = (*i)->GetPreds();
  }
  necUseful.insert(usefulLocs.begin(),usefulLocs.end());
}

/*********************************************************************/
//seed the control locations with necessary predicates
/*********************************************************************/
void ProcManager::ComputeNecessaryPredicates() const
{
  //add the predicate qualifying the initial state to the initial
  //location
  if(initLoc->GetPreds().AddPred(Predicate(guard,Predicate::REQUIRED,pair<const ContLoc*,int>(initLoc,1)))) {
    Util::Message(3,"predicate " + guard.ToString() + " added to control location " 
		  + initLoc->ToString() + "\n");
  }
  
  //add predicates associated with return actions to return locations
  AddRetActPreds();
  
  //add predicates associated with call actions to call locations
  AddCallActPreds();

  //add predicate from ltl formulas
  if(Database::CONF_TYPE == Database::CONF_LTL) AddLtlPreds();
}

/*********************************************************************/
//add the predicates associated with return actions to return
//locations
/*********************************************************************/
void ProcManager::AddRetActPreds() const
{
  //get the set of expresions associated with return actions
  set<Expr> exprs = Database::GetSpecRetActExprs();

  //convert the expressions to predicates. specifically an expression
  //expr becomes a predicate ($0 == expr).
  Expr dummy = ExprManager::GetIdExpr("$0");
  set<Expr> preds;
  for(set<Expr>::const_iterator i = exprs.begin();i != exprs.end();i++) {
    Expr a = ExprManager::GetBinaryExpr(dummy,*i,MAGIC_EQ_OP);
    preds.insert(a);
  }

  ProcInfo &procInfo = Database::procInfos[name];
  for(size_t i = 0;i < allLocs.size();i++) {
    ContLoc *a = allLocs[i];
    ProcInfo *b = a->GetProcInfo();
    //if this is not the info of the main procedure, there are
    //no return action predicates to add
    if((&procInfo) != b) continue;
    if(a->GetLocType() == ContLoc::RETURN) {
      a->AddRetActPreds(preds);
    }
  }
}

/*********************************************************************/
//add the predicates associated with call actions to call locations
/*********************************************************************/
void ProcManager::AddCallActPreds() const
{
  for(size_t i = 0;i < allLocs.size();i++) {
    ContLoc *loc = allLocs[i];
    //get the map of names of procedures in call actions to
    //set of predicates associated with the actions
    if(loc->GetLocType() == ContLoc::ASSIGN_CALL) {
      loc->AddCallActPreds();
    }
  }
}

/*********************************************************************/
//add the predicates associated with the spec ltl formula to all
//locations
/*********************************************************************/
void ProcManager::AddLtlPreds() const
{
  const set<Expr> &props = Database::procInfos[name].GetProps();
  for(size_t i = 0;i < allLocs.size();i++) {
    ContLoc *loc = allLocs[i];
    for(set<Expr>::const_iterator j = props.begin();j != props.end();j++) {
      if(loc->AddPred(Predicate(*j,Predicate::REQUIRED,pair<const ContLoc*,int>(loc,1)))) {
	Util::Message(3,"predicate " + j->ToString() + " added to control location " + loc->ToString() + "\n");
      }
    }
  }
}

/*********************************************************************/
//compute the set of implementation states for each control location
/*********************************************************************/
void ProcManager::ComputeImplStates()
{
  //display the seed branches
  Util::Message(3,"********* begin seed branches ***********\n");
  for(set<ContLoc*>::const_iterator i = seedBranches.begin();i != seedBranches.end();i++) {
    Util::Message(3,(*i)->ToString() + "\n");
  }
  Util::Message(3,"********* end seed branches ***********\n");

  //setup the locations for computing implementation states
  if(Database::CEGAR && Database::USE_PRED_ABSTRACTION) {
    SetupLocsCegar();
  } else {
    SetupLocsNoCegar();
  }

  //compute the spec infos
  for(size_t i = 0;i < usefulLocs.size();i++) {
    usefulLocs[i]->ComputeSpecInfos();
  }

  //compute the reachable implementation states
  reachImpl.clear();
  absLts.acts.clear();
  assert(Database::USE_REACH_IMPL);

  //first check the cache if necessary
  if(Database::CACHE_IMPL_TRANS) {
    if(seedMap.count(seedBranches) != 0) {
      char buf[64];
      snprintf(buf,64,"implstate-%d-%d-%d",getpid(),index,seedMap[seedBranches]);
      struct stat statStruct;
      if(stat(buf,&statStruct) == 0) {
	FILE *in = fopen(buf,"r");
	assert(in != NULL);
	//get the implementation states
	while(true) {
	  ContLoc *loc = NULL;
	  fscanf(in,"%p",&loc);
	  if(loc == 0x0) break;
	  unsigned long oldPVL;
	  short oldSS,oldSI;
	  fscanf(in,"%lu",&oldPVL);
	  fscanf(in,"%hd",&oldSS);
	  fscanf(in,"%hd",&oldSI);
	  PVal oldPV = PredSet::ULToPVal(oldPVL);
	  const ImplState *is = loc->GetImplState(oldPV,oldSS,oldSI);
	  reachImpl.insert(is);
	  set<Expr> props;
	  while(true) {
	    const BasicExpr *prop = NULL;
	    fscanf(in,"%p",&prop);
	    if(prop == NULL) break;
	    props.insert(Expr(prop));
	  }
	  loc->SetPropositions(is,props);
	}
	//get the actions
	while(true) {
	  short actId = Action::ACTION_NONE;
	  fscanf(in,"%hd",&actId);
	  if(actId == Action::ACTION_NONE) break;
	  absLts.acts.insert(Action(actId));
	}
	//all done
	fclose(in);
	return;
      }
    }
  }

  //if not in cache compute
  set<const ImplState*> front = initLoc->GetImplStates(guard);
  while(!front.empty()) {
    reachImpl.insert(front.begin(),front.end());
    set<const ImplState*> newFront = front;
    front.clear();
    for(set<const ImplState*>::const_iterator i = newFront.begin();i != newFront.end();i++) {	
      (*i)->GetContLoc()->SetPropositions(*i,ComputeTrueProps((*i)->GetContLoc(),(*i)->GetPredVal()));
      list< pair<const ImplState*,Action> > sa = (*i)->GetSuccsAndActions();
      for(list< pair<const ImplState*,Action> >::const_iterator k = sa.begin();k != sa.end();k++) {
	if(reachImpl.count(k->first) == 0) front.insert(k->first);
	absLts.acts.insert(k->second);
      }
    }
  }
  //no spec epsilon actions should be present
  assert(absLts.acts.count(ActionManager::GetSpecEpsilonAction()) == 0);
  //clear caches associated with predsets
  PredSet::ClearCache();

  //update the cache if necessary
  if(Database::CACHE_IMPL_TRANS) {
    int x = -1;
    if(seedMap.count(seedBranches) == 0) {
      x = seedMap.size();
      seedMap[seedBranches] = x;
    } else x = seedMap[seedBranches];
    char buf[64];
    snprintf(buf,64,"implstate-%d-%d-%d",getpid(),index,x);
    FILE *out = fopen(buf,"w");
    //output the reachable impl states
    for(set<const ImplState*>::const_iterator i = reachImpl.begin();i != reachImpl.end();i++) {
      fprintf(out,"%p %lu %hd %hd ",(*i)->GetContLoc(),PredSet::PValToUL((*i)->GetPredVal()),
	      (*i)->GetSpecStateId(),(*i)->GetSpecInfoId());
      set<Expr> props = (*i)->GetPropositions();
      for(set<Expr>::const_iterator j = props.begin();j != props.end();j++) {
	fprintf(out,"%p ",j->GetExpr());
      }
      fprintf(out,"0x0 ");
    }
    fprintf(out,"0x0 ");
    //output the actions
    for(set<Action>::const_iterator i = absLts.acts.begin();i != absLts.acts.end();i++) {
      fprintf(out,"%hd ",i->GetIndex());
    }
    fprintf(out,"%hd\n",Action::ACTION_NONE);
    //all done
    fclose(out);
  }
}

/*********************************************************************/
//setup the control locations as if we are doing cegar
/*********************************************************************/
void ProcManager::SetupLocsCegar()
{
  //clear pred sets at all locations
  ClearPredSets();

  //initialise the locations
  InitialiseLocs();

  //create the relevant prediactes by merging the maps of the seed
  //branch locations and necessary predicates
  for(set<ContLoc*>::const_iterator i = seedBranches.begin();i != seedBranches.end();i++) {
    for(vector<ContLoc*>::const_iterator j = allLocs.begin();j != allLocs.end();j++) {
      (*j)->GetPreds().Merge(locToPreds[*i][*j]);
    }
  }
  for(vector<ContLoc*>::const_iterator i = allLocs.begin();i != allLocs.end();i++) {
    (*i)->GetPreds().Merge(necPreds[*i]);
  }

  //compute the usefulness of control locations
  for(set<ContLoc*>::const_iterator i = seedBranches.begin();i != seedBranches.end();i++) {
    for(vector<ContLoc*>::const_iterator j = allLocs.begin();j != allLocs.end();j++) {
      if((*j)->GetLocType() == ContLoc::ASSIGN_EXPR) {
	if(locToUseful[*i].count(*j) != 0) {
	  (*j)->SetUseful(true);
	}
      }
    }
  }
  for(vector<ContLoc*>::const_iterator i = allLocs.begin();i != allLocs.end();i++) {
    if((*i)->GetLocType() == ContLoc::ASSIGN_EXPR) {
      if(necUseful.count(*i) != 0) {
	(*i)->SetUseful(true);
      }
    }
  }
  
  //remove useless control locations and recreate the control
  //location array
  RemoveUselessContLocs();
}

/*********************************************************************/
//setup the control locations as if we are not doing cegar
/*********************************************************************/
void ProcManager::SetupLocsNoCegar()
{
  //initialise the locations
  InitialiseLocs();

  //compute the set of predicates relevant for each control
  //location
  ComputeRelevantPreds();
  Util::Message(2,"relevant predicates computed ...\n");
  
  //remove useless control locations and recreate the control
  //location array
  RemoveUselessContLocs();
  Util::Message(2,"useless locations eliminated ...\n");
}

/*********************************************************************/
//initialise stuff before seting up control locations
/*********************************************************************/
void ProcManager::InitialiseLocs() const
{
  //restore all important fields of the control locations,cleanup
  //implementation states, specification state names, successor
  //implementation states and simulation relation and set the useful
  //fields to false.
  for(size_t i = 0;i < allLocs.size();i++) {
    allLocs[i]->Restore();
    if(allLocs[i]->GetLocType() == ContLoc::ASSIGN_EXPR) {
      allLocs[i]->SetUseful(false);
    }
  }
}

/*********************************************************************/
//clear predicate sets at all control locations
/*********************************************************************/
void ProcManager::ClearPredSets()
{ 
  for(size_t i = 0;i < allLocs.size();i++) {
    allLocs[i]->GetPreds().Clear();
  }
}

/*********************************************************************/
//compute the set of predicates relevant for each location. the first
//argument is the call graph. the second argument is the info about
//the main procedure. the third argument is the predicate qualifying
//the initial state. also compute usefulness of locations w.r.t. the
//computed set of predicates.
/*********************************************************************/
void ProcManager::ComputeRelevantPreds()
{ 
  //initialise the working set
  set<ContLoc*> wset;
  for(size_t i = 0;i < allLocs.size();i++) {
    wset.insert(allLocs[i]);
  }
  
  //compute the relevant predicates
  while(!wset.empty()) {
    ContLoc *a = *(wset.begin());
    wset.erase(wset.begin());
    if(a->UpdatePreds()) {
      const set<ContLoc*> &b = a->GetPredecessors();
      wset.insert(b.begin(),b.end());
    }
  }

  //compute the usefulness of the locations
  for(size_t i = 0;i < allLocs.size();i++) {
    allLocs[i]->ComputeUseful();
  }
}

/*********************************************************************/
//remove useless assign expression locations and branch locations
//whose true and false successors are the same. finally recreate the
//global control location array.
/*********************************************************************/
void ProcManager::RemoveUselessContLocs()
{
  //remove useless locations
  if(!initSkipped.empty()) {
    initLoc = initSkipped.front();
    initSkipped.clear();
  }
  list<ContLoc*> contLocs;
  for(size_t i = 0;i < allLocs.size();i++) {
    allLocs[i]->GetTSkipped().clear();
    allLocs[i]->GetESkipped().clear();
    allLocs[i]->GetCallSkipped().clear();
    contLocs.push_back(allLocs[i]);	
  }

  size_t removed = 0;
  if(Database::NO_USELESS_LOCS) {
    bool flag = false;
    do {
      flag = false;
      for(list<ContLoc*>::iterator i = contLocs.begin();i != contLocs.end();) {
	ContLoc *a = *i;
	if((a != a->GetTSucc()) && (a != initLoc) &&
	   (((a->GetLocType() == ContLoc::ASSIGN_EXPR)&&(!a->IsUseful())) ||
	    ((a->GetLocType() == ContLoc::BRANCH) && (a->GetTSucc() == a->GetESucc())))) {
	  for(list<ContLoc*>::iterator j = contLocs.begin();j != contLocs.end();j++) {
	    if(a == *j) continue;
	    (*j)->ReplaceSucc(a,a->GetTSucc());
	  }
	  if(initLoc == a) {
	    initSkipped.push_back(a);
	    initLoc = a->GetTSucc();
	  }
	  i = contLocs.erase(i);
	  removed++;
	  flag = true;
	} else {
	  i++;
	}
      }
    } while(flag);
  }

  //create the useful location vector
  usefulLocs.clear();
  size_t id = 0;
  for(list<ContLoc*>::iterator i = contLocs.begin();i != contLocs.end();i++,id++) {
    usefulLocs.push_back(*i);
    (*i)->SetLocId(id * Database::rootProcNames.size() + index);
  }
  assert(usefulLocs.size() == (allLocs.size() - removed));
  Util::Message(4,"array of useful control locations created ...\n");
}

/*********************************************************************/
//compute the successors and actions of control locations
/*********************************************************************/
void ProcManager::ComputeSuccsAndActions()
{
  //first check the cache if necessary
  if(Database::CACHE_IMPL_TRANS) {
    if(seedMap.count(seedBranches) != 0) {
      char buf[64];
      snprintf(buf,64,"impltrans-%d-%d-%d",getpid(),index,seedMap[seedBranches]);
      struct stat statStruct;
      if(stat(buf,&statStruct) == 0) {
	FILE *in = fopen(buf,"r");
	while(true) {
	  ContLoc *loc = NULL;
	  fscanf(in,"%p",&loc);
	  if(loc == 0x0) break;
	  Util::Message(4,"restoring succs and actions for " + loc->ToString() + 
			" with predicates : " + loc->GetPreds().ToString() + "\n");
	  loc->RestoreSuccsAndActions(in);
	}
	fclose(in);
	return;
      }
    }
  }

  //compute the successors and actions of every control location
  for(size_t i = 0;i < usefulLocs.size();i++) {
    Util::Message(2,"computing succs and actions for " + usefulLocs[i]->ToString() + 
		  " with predicates : " + usefulLocs[i]->GetPreds().ToString() + "\n");
    usefulLocs[i]->ComputeSuccsAndActions();
  }
  PredSet::ClearCache();

  //update the cache if necessary
  if(Database::CACHE_IMPL_TRANS) {
    int x = -1;
    if(seedMap.count(seedBranches) == 0) {
      x = seedMap.size();
      seedMap[seedBranches] = x;
    } else x = seedMap[seedBranches];
    char buf[64];
    snprintf(buf,64,"impltrans-%d-%d-%d",getpid(),index,x);
    FILE *out = fopen(buf,"w");
    for(size_t i = 0;i < usefulLocs.size();i++) {
      Util::Message(4,"backing up succs and actions for " + usefulLocs[i]->ToString() + 
		    " with predicates : " + usefulLocs[i]->GetPreds().ToString() + "\n");
      fprintf(out,"%p ",usefulLocs[i]);
      usefulLocs[i]->BackupSuccsAndActions(out);
    }
    fprintf(out,"0x0\n");
    fclose(out);
  }
}

/*********************************************************************/
//create the initial abstract states by grouping together
//implementation states with the same set of outgoing actions
/*********************************************************************/
void ProcManager::CreateInitialAbstractStates()
{
  absLts.stateNum = 0;
  absLts.propMap.clear();
  map< pair< set<Expr>,set<Action> >,size_t > propActCombs;
  for(set<const ImplState*>::const_iterator i = reachImpl.begin();i != reachImpl.end();i++) {
    ImplState *istate = (ImplState*)(*i);
    //if we need to use lts abstraction
    if(Database::USE_LTS_ABSTRACTION) {
      pair< set<Expr>,set<Action> > propActs;
      propActs.first = istate->GetPropositions();
      list< pair<const ImplState*,Action> > sa = istate->GetSuccsAndActions();
      for(list< pair<const ImplState*,Action> >::const_iterator k = sa.begin();k != sa.end();k++) {
	propActs.second.insert(k->second);
      }
      
      map< pair< set<Expr>,set<Action> >,size_t >::iterator it = propActCombs.find(propActs);
      if(it == propActCombs.end()) {
	propActCombs[propActs] = absLts.stateNum;
	istate->SetAbsId(absLts.stateNum);
	absLts.propMap[absLts.stateNum] = propActs.first;
	absLts.stateNum++;
      } else {
	istate->SetAbsId(it->second);
      }
    }
    //otherwise each abstract states contains just a single
    //implementation state
    else {
      istate->SetAbsId(absLts.stateNum);      
      absLts.propMap[absLts.stateNum] = istate->GetPropositions();
      absLts.stateNum++;
    }
  }
}

/*********************************************************************/
//compute abstract initial states and actions
/*********************************************************************/
void ProcManager::ComputeAbsLtsInit()
{
  Util::Message(2,"number of abstract implementation states = %d\n",absLts.stateNum);
  //compute initial states
  absLts.init.clear();
  set<const ImplState*> initImpls = initLoc->GetImplStates(guard);
  for(set<const ImplState*>::const_iterator i = initImpls.begin();i != initImpls.end();i++) {
    absLts.init.insert((*i)->GetAbsId());
  }
  //clear the transitions
  absLts.succMap.clear();
  absLts.predMap.clear();
}

/*********************************************************************/
//compute abstract transitions
/*********************************************************************/
void ProcManager::ComputeAbsLtsTrans()
{
  for(set<const ImplState*>::const_iterator i = reachImpl.begin();i != reachImpl.end();i++) {
    list< pair<const ImplState*,Action> > sa = (*i)->GetSuccsAndActions();
    for(list< pair<const ImplState*,Action> >::const_iterator k = sa.begin();k != sa.end();k++) {
      absLts.succMap[(*i)->GetAbsId()][k->second].insert(k->first->GetAbsId());
      absLts.predMap[k->first->GetAbsId()][k->second].insert((*i)->GetAbsId());
    }
  }
}

/*********************************************************************/
//given a control location and a predicate valuation compute the set
//of true propositions in the corresponding impl state
/*********************************************************************/
set<Expr> ProcManager::ComputeTrueProps(const ContLoc *loc,const PVal &pv) const
{
  set<Expr> res;
  if(Database::CONF_TYPE == Database::CONF_LTL) {
    set<Expr> state = loc->GetPreds().ToExpr(loc->GetPreds().PValToPredVal(pv));
    const set<Expr> &props = Database::procInfos[name].GetProps();
    for(set<Expr>::const_iterator i = props.begin();i != props.end();i++) {
      bool assigned = false;
      for(set<Expr>::const_iterator j = state.begin();j != state.end();j++) {
	if(ExprManager::Implies(*j,*i)) {
	  res.insert(*i);
	  assigned = true;
	  break;
	} else if(ExprManager::DisjointWith(*j,*i)) {
	  assigned = true;
	  break;
	}
      }
      assert(assigned);
    }
  }
  return res;
}

/*********************************************************************/
//draw the control flow graph as a dot file
/*********************************************************************/
void ProcManager::DrawCFG() const
{
  FILE *out = Database::dotFilePtr;
  fprintf(out,"  //CFG of Component %s index %d\n",name.c_str(),index);
  fprintf(out,"  subgraph cluster_cfg_%s_%d {\n",name.c_str(),index);
  fprintf(out,"    color = white;\n");
  fprintf(out,"    cont_loc_%p [peripheries = 2];\n",initLoc);   
  for(vector<ContLoc*>::const_iterator i = allLocs.begin();i != allLocs.end();i++) {
    fprintf(out,"    cont_loc_%p [label = \"%s\"];\n",(*i),(*i)->ToString().c_str());    
    if((*i)->GetLocType() == ContLoc::BRANCH) {
      fprintf(out,"    cont_loc_%p -> cont_loc_%p [label = \"then\"];\n",*i,(*i)->GetTSucc());
      fprintf(out,"    cont_loc_%p -> cont_loc_%p [label = \"else\"];\n",*i,(*i)->GetESucc());
    } else if((*i)->GetLocType() != ContLoc::FINAL){
      fprintf(out,"    cont_loc_%p -> cont_loc_%p;\n",*i,(*i)->GetTSucc());
    }
  }
  fprintf(out,"    label = \"CFG of component #%d \\nProcedure name: %s\";\n",index,name.c_str());
  fprintf(out,"    labelloc = \"b\";\n");
  fprintf(out,"  }\n\n");
}

/*********************************************************************/
//draw the control flow graph with inferred predicates
/*********************************************************************/
void ProcManager::DrawInferredPreds() const
{
  FILE *out = Database::dotFilePtr;
  fprintf(out,"  //Inferred predicates of component %s index %d\n",name.c_str(),index);
  fprintf(out,"  subgraph cluster_inf_preds_%s_%d_%d {\n",name.c_str(),index,Statistics::iterNum);
  fprintf(out,"    color = white;\n");
  for(vector<ContLoc*>::const_iterator i = allLocs.begin();i != allLocs.end();i++) {
    string recStr = Util::EscapeCharacter((*i)->ToString(),'|');    
    const vector<Predicate> &preds = (*i)->GetPreds().GetPreds();
    for(vector<Predicate>::const_iterator j = preds.begin();j != preds.end();j++) {
      for(vector<Expr>::const_iterator k = j->GetValToExpr().begin();k != j->GetValToExpr().end();k++) {
	recStr += " | " + Util::TrimString(Util::EscapeCharacter(k->ToString(),'|'));
      }
    }
    recStr = Util::EscapeCharacter(recStr,'<');    
    recStr = Util::EscapeCharacter(recStr,'>');    
    recStr = Util::EscapeCharacter(recStr,'{');    
    recStr = Util::EscapeCharacter(recStr,'}');        
    fprintf(out,"    inf_preds_%p_%d [shape=Mrecord,label = \"{%s}\"];\n",(*i),Statistics::iterNum,recStr.c_str());    
    if((*i)->GetLocType() == ContLoc::BRANCH) {
      fprintf(out,"    inf_preds_%p_%d -> inf_preds_%p_%d [label = \"then\"];\n",*i,Statistics::iterNum,(*i)->GetTSucc(),Statistics::iterNum);
      fprintf(out,"    inf_preds_%p_%d -> inf_preds_%p_%d [label = \"else\"];\n",*i,Statistics::iterNum,(*i)->GetESucc(),Statistics::iterNum);
    } else if((*i)->GetLocType() != ContLoc::FINAL){
      fprintf(out,"    inf_preds_%p_%d -> inf_preds_%p_%d;\n",*i,Statistics::iterNum,(*i)->GetTSucc(),Statistics::iterNum);
    }
  }
  fprintf(out,"    label = \"Inferred predicates \\nComponent #%d \\nProcedure name: %s\\nIteration #%d\";\n",index,name.c_str(),Statistics::iterNum);
  fprintf(out,"    labelloc = \"b\";\n");
  fprintf(out,"  }\n\n");
}

/*********************************************************************/
//output the implementation machine obtained by predicate abstraction
//as a dot file
/*********************************************************************/
void ProcManager::DrawPredAbsLts() const
{
  FILE *out = Database::dotFilePtr;
  fprintf(out,"  //LTS obtained by predicate abstraction of component %s index %d\n",name.c_str(),index);
  fprintf(out,"  subgraph cluster_pred_abs_lts_%s_%d_%d {\n",name.c_str(),index,Statistics::iterNum);
  fprintf(out,"    color = white;\n");
  string tv = Util::NewTempVar();
  fprintf(out,"    %s [label = \"INIT\",peripheries = 2];\n",tv.c_str());
  set<const ImplState*> front = initLoc->GetImplStates(guard);
  for(set<const ImplState*>::const_iterator i = front.begin();i != front.end();i++) {
    fprintf(out,"    %s -> pred_abs_lts_%p_%d [label =\"epsilon_%d\"];\n",tv.c_str(),*i,Statistics::iterNum,index);
  }  
  for(set<const ImplState*>::const_iterator i = reachImpl.begin();i != reachImpl.end();i++) {
    const ContLoc *loc = (*i)->GetContLoc();
    const PredSet &preds = loc->GetPreds();
    string recStr = Util::EscapeCharacter(loc->ToString(),'|');    
    PredVal pval = preds.PValToPredVal((*i)->GetPredVal());
    set<Expr> eset = preds.ToExpr(pval);
    for(set<Expr>::const_iterator j = eset.begin();j != eset.end();j++) {
      recStr += " | " + Util::TrimString(Util::EscapeCharacter(j->ToString(),'|'));
    }
    int ssi = (*i)->GetSpecStateId();
    if(ssi != -1) {
      recStr += " | " + Database::ltsInfos[(*i)->GetSpecInfoId()].GetIdState(ssi);
    }
    recStr = Util::EscapeCharacter(recStr,'<');    
    recStr = Util::EscapeCharacter(recStr,'>');    
    recStr = Util::EscapeCharacter(recStr,'{');    
    recStr = Util::EscapeCharacter(recStr,'}');        
    fprintf(out,"    pred_abs_lts_%p_%d [shape=Mrecord,label = \"{%s}\"];\n",*i,Statistics::iterNum,recStr.c_str());    
    list< pair<const ImplState*,Action> > saa = (*i)->GetSuccsAndActions();
    for(list< pair<const ImplState*,Action> >::const_iterator j = saa.begin();j != saa.end();j++) {
      fprintf(out,"    pred_abs_lts_%p_%d -> pred_abs_lts_%p_%d [label = \"%s\"];\n",*i,Statistics::iterNum,j->first,Statistics::iterNum,j->second.ToString().c_str());
    }
  }
  fprintf(out,"    label = \"LTS after predicate abstraction \\nComponent #%d \\nProcedure name: %s\\nIteration #%d\";\n",index,name.c_str(),Statistics::iterNum);
  fprintf(out,"    labelloc = \"b\";\n");
  fprintf(out,"  }\n\n");
}

/*********************************************************************/
//output the model created by predicate abstraction
/*********************************************************************/
void ProcManager::OutputModel() const
{
  FILE *out = Database::modelFilePtr;
  if(Database::OUTPUT_MODEL_TYPE == Database::OUTPUT_MODEL_FSP) {
    fprintf(out,"//Iteration #%d, Component #%d, Procedure name: %s\n",Statistics::iterNum,index,name.c_str());
    set<const ImplState*> front = initLoc->GetImplStates(guard);
    for(set<const ImplState*>::const_iterator i = front.begin();i != front.end();i++) {
      fprintf(out,"INIT --> epsilon_%d --> %s\n",index,(*i)->ToString().c_str());
    }
    for(set<const ImplState*>::const_iterator i = reachImpl.begin();i != reachImpl.end();i++) {
      list< pair<const ImplState*,Action> > saa = (*i)->GetSuccsAndActions();
      for(list< pair<const ImplState*,Action> >::const_iterator j = saa.begin();j != saa.end();j++) {
	fprintf(out,"%s --> %s --> %s\n",(*i)->ToString().c_str(),j->second.ToString().c_str(),j->first->ToString().c_str());	
      }
    }
    fprintf(out,"\n");
  } else assert(false);
}

/*********************************************************************/
//end of ProcManager.cpp
/*********************************************************************/
