/**************************** CPPHeaderFile ***************************

* FileName [ProcManager.h]

* PackageName [main]

* Synopsis [This class encapsulates a procedure manager.]

* Description []

* SeeAlso [arc.cpp]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University.  All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#ifndef __PROCMANAGER_H__
#define __PROCMANAGER_H__

//other classes needed
class ContLoc;
class CallGraph;
class ConcCEDag;

/*********************************************************************/
//the procedure manager
/*********************************************************************/
class ProcManager
{
 private:
  //the index of this manager in the global array
  size_t index;

  //the name of the procedure
  string name;

  //the abstraction guard
  Expr guard;

  //the array of all control locations after eliminating skip
  //locations
  vector<ContLoc*> allLocs;

  //the set of all branch locations
  set<ContLoc*> branchLocs;

  //the array of useful control locations according to the predicates
  //being used in a particular refinement cycle
  vector<ContLoc*> usefulLocs;

  //the initial control location
  ContLoc *initLoc;

  //the list of initially skipped control locations. these locations
  //are eliminated since they are useless w.r.t. the set of predicates
  //being considered
  list<ContLoc*> initSkipped;

  //the final control location
  ContLoc *finalLoc;

  //the set of reachable implementation states
  set<const ImplState*> reachImpl;

  //the traces over which we are optimizing
  set<ConcCEDag> optDags;

  //the map from branch locations to predicates to install
  map< ContLoc*,map<ContLoc*,PredSet> > locToPreds;

  //the map from branch locations to useful locations
  map< ContLoc*,set<ContLoc*> > locToUseful;

  //the map from control locations to necessary predicates
  map<ContLoc*,PredSet> necPreds;

  //set of useful locations due to necessary predicates
  set<ContLoc*> necUseful;

  //the map from assignment locations to sets of branch
  //locations. location X is mapped to Y if some predicate inferred
  //from Y is related to the constraint created by X
  map< Expr,set<ContLoc*> > assignToBranch;

  //the set of branch locations whose branch conditions are used for
  //predicate seeding
  set<ContLoc*> seedBranches;

  //a flag indicating if the seed branches have been modified
  bool newSeeds;

  //the local abstract lts
  LocalAbsLts absLts;

  //the map of branches to indices
  map<ContLoc*, int> varMap;

  //the map of indices to branches
  map<int, ContLoc*> mapVar;

  //the set of all constraints in CNF form
  set< set<ContLoc*> > allCons;

  //map from spurious dags discovered to control location sets that
  //can possibly rule it out
  map< ConcCEDag,set<ContLoc*> > dagToLocs;

  //the number of clauses generated and stored in the constraint file
  size_t clauseNum;

  //the name of the constraints file. this contains just a list of the
  //disjunctive clauses. this gets updated in every iteration. the
  //dimacs file is constructed from this file.
  string constraintFile;

  //the name of the dimacs file
  string dimacsFile;

  //the name of the pb file
  string pbFile;

  //the name of the output file
  string outputFile;

  //the map of seeds cached to their indices
  map< set<ContLoc*>,int > seedMap;

  //the state of the abstract local lts that should be split and the
  //action on which it should be split
  map< int,set<Action> > absLtsRefine;

  void AddRetActPreds() const;
  void AddCallActPreds() const;
  void AddLtlPreds() const;

  void InitialiseLocs() const;
  void SetupLocsCegar();
  void ComputeRelevantPreds();
  void RemoveUselessContLocs();

  void SetSeedBranches(const set<ContLoc*> &s) { seedBranches = s; newSeeds = true; }
  set<Expr> ComputeTrueProps(const ContLoc *loc,const PVal &pv) const;
  void DrawCFG() const;

 public:
  ProcManager();
  ProcManager(size_t i,const string &n,const Expr &g);
  ProcManager(const ProcManager &rhs);

  const ProcManager &operator = (const ProcManager &rhs);

  void Initialise() const;
  void Cleanup() const;

  const Expr &GetGuard() const { return guard; }
  const vector<ContLoc*> &GetAllLocs() const { return allLocs; }
  const vector<ContLoc*> &GetUsefulLocs() const { return usefulLocs; }
  ContLoc *GetInitLoc() { return initLoc; }
  list<ContLoc*> &GetInitSkipped() { return initSkipped; }
  ContLoc *GetFinalLoc() { return finalLoc; }
  const set<const ImplState*> &GetReachImpl() const { return reachImpl; }
  const set<ContLoc*> &GetSeedBranches() const { return seedBranches; }
  const LocalAbsLts &GetAbsLts() const { return absLts; }
  LocalAbsLts &GetAbsLts() { return absLts; }
  bool GetNewSeeds() const { return newSeeds; }

  void SetNewSeeds(const bool n) { newSeeds = n; }

  void CreateAllContLocs();
  void DoLiveVariableAnalysis() const;
  void BuildInitialMaps();
  void ComputeNecessaryPredicates() const;

  void SetupLocsNoCegar();
  void ClearPredSets();
  void ComputeImplStates();
  void ComputeSuccsAndActions();
  void CreateInitialAbstractStates();
  void ComputeAbsLtsInit();
  void ComputeAbsLtsTrans();

  void DrawInferredPreds() const;
  void DrawPredAbsLts() const;
  void OutputModel() const;

  friend class CEDagVerifier;
  friend class AbsRefiner;
  friend class PredAbsRefiner;
  friend class LtsAbsRefiner;
  friend class CEGen;
  friend class ConcCEDag;
};

#endif //__PROCMANAGER_H__

/*********************************************************************/
//end of ProcManager.h
/*********************************************************************/
