/******************************** CPPFile *****************************

* FileName [Action.cpp]

* PackageName [parser]

* Synopsis [Method definitions of Action class.]

* SeeAlso [Action.h]

* Author [Sagar Chaki]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University. All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: chaki+@cs.cmu.edu. ]

**********************************************************************/

#include <cstdio>
#include <cassert>
#include <string>
#include <list>
#include <map>
#include <set>
#include <vector>
using namespace std;

#include "Node.h"
#include "Action.h"
#include "ActionManager.h"
#include "Database.h"
#include "StdcParser.h"

/*********************************************************************/
//define static members
/*********************************************************************/
const int Action::ACTION_NONE = -10;

/*********************************************************************/
//constructors
/*********************************************************************/
//creating from an id
Action::Action(short i)
{
  assert(i >= 0);
  index = i;
}

//for creating a basic action
Action::Action(const string &n) 
{
  index = ActionManager::RegisterBasicAction(n);
}

//for creating a return action
Action::Action(const Expr &e) 
{
  index = ActionManager::RegisterReturnAction(e);
}

//for creating an assign action
Action::Action(const Expr &e1,const Expr &e2) 
{
  index = ActionManager::RegisterAssignAction(e1,e2);
}

/*********************************************************************/
//operators
/*********************************************************************/
const Action &Action::operator = (const Action &rhs) 
{
  index = rhs.index;
  return *this;
}

bool Action::operator == (const Action &rhs) const
{
  return (index == rhs.index);
}

//for the less than operator we first compare the types and then the
//string representations
bool Action::operator < (const Action &rhs) const
{
  return (index < rhs.index);
}

/*********************************************************************/
//accessors
/*********************************************************************/
string Action::GetName() const
{
  return ActionManager::GetActions()[index].GetName();
}

const Expr &Action::GetRetExpr() const
{
  return ActionManager::GetActions()[index].GetRetExpr();
}

const Expr &Action::GetLhsExpr() const
{
  return ActionManager::GetActions()[index].GetLhsExpr();
}

const Expr &Action::GetRhsExpr() const
{
  return ActionManager::GetActions()[index].GetRhsExpr();
}

/*********************************************************************/
//convert to string format
/*********************************************************************/
string Action::ToString() const
{
  return ActionManager::GetActions()[index].ToString();
}

/*********************************************************************/
//returns true if this is an spec epsilon action and false otherwise
/*********************************************************************/
bool Action::IsSpecEpsilon() const
{
  const map<int,short>::const_iterator i = ActionManager::GetEpsilonIndices().find(-1);
  if(i == ActionManager::GetEpsilonIndices().end()) return false;
  else return (index == i->second);
}

/*********************************************************************/
//returns true if this is a default return action and false otherwise
/*********************************************************************/
bool Action::IsDefaultRet() const
{
  return (index == ActionManager::GetDefRetIndex());
}

/*********************************************************************/
//returns true if this is a basic action and false otherwise
/*********************************************************************/
bool Action::IsBasic() const
{
  return ActionManager::GetActions()[index].IsBasic();
}

/*********************************************************************/
//returns true if this is a return action and false otherwise
/*********************************************************************/
bool Action::IsReturn() const
{
  return ActionManager::GetActions()[index].IsReturn();
}

/*********************************************************************/
//returns true if this is an assign action and false otherwise
/*********************************************************************/
bool Action::IsAssign() const
{
  return ActionManager::GetActions()[index].IsAssign();
}

/*********************************************************************/
//return true if this is a wildcard return and false otherwise
/*********************************************************************/
bool Action::IsWildcardRet() const
{
  return ActionManager::GetActions()[index].IsWildcardRet();
}

/*********************************************************************/
//given return expressions, the parameter list of a procedure and a
//context, return true if the return statement can cause the return
//action and false otherwise
/*********************************************************************/
bool Action::ReturnActionPossible(const Expr &retExpr,const list<string> &params,
				  const set<Expr> &context) const
{
  return ActionManager::GetActions()[index].ReturnActionPossible(retExpr,params,context);
}

/*********************************************************************/
//given expressions for the lhs and rhs, the parameter list of a
//procedure and a context, return true if the assignment statement can
//cause the assignment action and false otherwise
/*********************************************************************/
bool Action::AssignActionPossible(const Expr &lhs,const Expr &rhs,
				  const list<string> &params,const set<Expr> &context) const
{
  return ActionManager::GetActions()[index].AssignActionPossible(lhs,rhs,params,context);
}

/*********************************************************************/
//return the concrete expression for this action by replacing "$1",
//"$2" etc with the list of arguments provided as argument.
/*********************************************************************/
Expr Action::GetConcreteReturnExpr(const list<Expr> &args) const
{
  return ActionManager::GetActions()[index].GetConcreteReturnExpr(args);
}

/*********************************************************************/
//return the pair of concrete expression for the lhs and rhs of this
//action by replacing "$1", "$2" etc with the list of arguments
//provided as argument.
/*********************************************************************/
pair<Expr,Expr> Action::GetConcreteAssignExprs(const list<Expr> &args) const
{
  return ActionManager::GetActions()[index].GetConcreteAssignExprs(args);
}

/*********************************************************************/
//end of Action.cpp
/*********************************************************************/
